import functools
import logging
import operator

from haoda import ir
from haoda import util
from soda import core
from soda import util as soda_util

logger = logging.getLogger().getChild(__name__)


def print_header(printer):

  # C headers
  for header in ('assert', 'float', 'math', 'stdbool', 'stddef', 'stdint',
                 'stdio', 'stdlib', 'string'):
    printer.println('#include <c%s>' % header)
  printer.println()

  # C++ headers
  for header in ('algorithm', 'array', 'random', 'string', 'unordered_map'):
    printer.println('#include <%s>' % header)
  printer.println()

  # Other system headers
  for header in ('ap_int', 'fcntl', 'time', 'unistd', 'sys/types', 'sys/stat',
                 'CL/opencl'):
    printer.println('#include <%s.h>' % header)
  printer.println()

  for name in ('std::array', 'std::count', 'std::fill', 'std::string',
               'std::unordered_map'):
    printer.println('using %s;' % name)
  printer.println()


def print_load_xclbin2(printer):
  println = printer.println
  do_scope = printer.do_scope
  un_scope = printer.un_scope

  def invalid_xclbin2():
    println('*kernel_binary = nullptr;')
    println('fprintf(*error_report, '
            '"ERROR: %s is not a valid xclbin2 file\\n", filename);')
    println('return -2;')

  println('FILE* const* error_report = &stderr;')
  println()
  println('int64_t load_xclbin2_to_memory(const char *filename, '
          'char **kernel_binary, char** device)')
  do_scope()
  println('uint64_t size = 0;')
  println('FILE *f = fopen(filename, "rb");')
  println('if(nullptr == f)')
  do_scope()
  println('*kernel_binary = nullptr;')
  println('fprintf(*error_report, "ERROR: cannot open %s\\n", filename);')
  println('return -1;')
  un_scope()
  println('char magic[8];')
  println('unsigned char cipher[32];')
  println('unsigned char key_block[256];')
  println('uint64_t unique_id;')
  with printer.if_('fread(magic, sizeof(magic), 1, f) != 1'):
    invalid_xclbin2()
  with printer.if_('strcmp(magic, "xclbin2") != 0'):
    invalid_xclbin2()
  with printer.if_('fread(cipher, sizeof(cipher), 1, f) != 1'):
    invalid_xclbin2()
  with printer.if_('fread(key_block, sizeof(key_block), 1, f) != 1'):
    invalid_xclbin2()
  with printer.if_('fread(&unique_id, sizeof(unique_id), 1, f) != 1'):
    invalid_xclbin2()
  with printer.if_('fread(&size, sizeof(size), 1, f) != 1'):
    invalid_xclbin2()
  println('char* p = new char[size+1]();')
  println('*kernel_binary = p;')
  println('memcpy(p, magic, sizeof(magic));')
  println('p += sizeof(magic);')
  println('memcpy(p, cipher, sizeof(cipher));')
  println('p += sizeof(cipher);')
  println('memcpy(p, key_block, sizeof(key_block));')
  println('p += sizeof(key_block);')
  println('memcpy(p, &unique_id, sizeof(unique_id));')
  println('p += sizeof(unique_id);')
  println('memcpy(p, &size, sizeof(size));')
  println('p += sizeof(size);')
  println('uint64_t size_left = size - sizeof(magic) - sizeof(cipher) - '
          'sizeof(key_block) - sizeof(unique_id) - sizeof(size);')
  println('if(size_left != fread(p, sizeof(char), size_left, f))')
  do_scope()
  println('delete[] p;')
  println('fprintf(*error_report, "ERROR: %s is corrupted\\n", filename);')
  println('return -3;')
  un_scope()
  println('*device = p + 5*8;')
  println('fclose(f);')
  println('return size;')
  un_scope()
  println()


def print_halide_rewrite_buffer(printer):
  println = printer.println
  do_scope = printer.do_scope
  un_scope = printer.un_scope

  println('static bool halide_rewrite_buffer(buffer_t *b, int32_t elem_size,')
  for dim in range(4):
    println('                  int32_t min{0}, int32_t extent{0}, '
            'int32_t stride{0}{1}'.format(dim, ',,,)' [dim]))
  do_scope()
  for item in 'min', 'extent', 'stride':
    for dim in range(4):
      println('b->{0}[{1}] = {0}{1};'.format(item, dim))
  println('return true;')
  un_scope()
  println()


def print_halide_error_codes(printer):
  i = 0
  for item in [
      'success', 'generic_error', 'explicit_bounds_too_small', 'bad_elem_size',
      'access_out_of_bounds', 'buffer_allocation_too_large',
      'buffer_extents_too_large', 'constraints_make_required_region_smaller',
      'constraint_violated', 'param_too_small', 'param_too_large',
      'out_of_memory', 'buffer_argument_is_null', 'debug_to_file_failed',
      'copy_to_host_failed', 'copy_to_device_failed', 'device_malloc_failed',
      'device_sync_failed', 'device_free_failed', 'no_device_interface',
      'matlab_init_failed', 'matlab_bad_param_type', 'internal_error',
      'device_run_failed', 'unaligned_host_ptr', 'bad_fold',
      'fold_factor_too_small'
  ]:
    printer.println('int halide_error_code_%s = %d;' % (item, i))
    i -= 1
  printer.println()


def print_halide_error_report(printer):
  println = printer.println
  println('int halide_error_bad_elem_size(void *user_context, '
          'const char *func_name,')
  println('                 const char *type_name, int elem_size_given, '
          'int correct_elem_size) {')
  println('  fprintf(*error_report, "%s has type %s but elem_size of the '
          'buffer passed in is %d instead of %d",')
  println('      func_name, type_name, elem_size_given, correct_elem_size);')
  println('  return halide_error_code_bad_elem_size;')
  println('}')
  println('int halide_error_constraint_violated(void *user_context, '
          'const char *var, int val,')
  println('                   const char *constrained_var, '
          'int constrained_val) {')
  println('  fprintf(*error_report, "Constraint violated: %s (%d) == %s (%d)",')
  println('      var, val, constrained_var, constrained_val);')
  println('  return halide_error_code_constraint_violated;')
  println('}')
  println('int halide_error_buffer_allocation_too_large(void *user_context, '
          'const char *buffer_name, uint64_t allocation_size, '
          'uint64_t max_size) {')
  println('  fprintf(*error_report, "Total allocation for buffer %s is %lu, '
          'which exceeds the maximum size of %lu",')
  println('      buffer_name, allocation_size, max_size);')
  println('  return halide_error_code_buffer_allocation_too_large;')
  println('}')
  println('int halide_error_buffer_extents_too_large(void *user_context, '
          'const char *buffer_name, int64_t actual_size, int64_t max_size) {')
  println('  fprintf(*error_report, "Product of extents for buffer %s is %ld, '
          'which exceeds the maximum size of %ld",')
  println('      buffer_name, actual_size, max_size);')
  println('  return halide_error_code_buffer_extents_too_large;')
  println('}')
  println('int halide_error_access_out_of_bounds(void *user_context, '
          'const char *func_name, int dimension, int min_touched, '
          'int max_touched, int min_valid, int max_valid) {')
  println('  if(min_touched < min_valid) {')
  println('    fprintf(*error_report, "%s is accessed at %d, which is before '
          'the min (%d) in dimension %d", func_name, min_touched, min_valid, '
          'dimension);')
  println('  } else if(max_touched > max_valid) {')
  println('    fprintf(*error_report, "%s is acccessed at %d, which is beyond '
          'the max (%d) in dimension %d", func_name, max_touched, max_valid, '
          'dimension);')
  println('  }')
  println('  return halide_error_code_access_out_of_bounds;')
  println('}')
  println()


# pylint: disable=too-many-branches,too-many-statements
def print_wrapped(printer, stencil):
  println = printer.println
  do_indent = printer.do_indent
  un_indent = printer.un_indent
  do_scope = printer.do_scope
  un_scope = printer.un_scope
  new_var = printer.new_var
  last_var = printer.last_var
  tensors = [(_.name, _.c_type) for _ in stencil.input_stmts] + \
            [(_.name, _.c_type) for _ in stencil.output_stmts] + \
            [(_.name, _.c_type) for _ in stencil.param_stmts]
  println('static int %s_wrapped(%sconst char* xclbin) HALIDE_FUNCTION_ATTRS' %
          (stencil.app_name, ''.join(
              [('buffer_t *var_%s_buffer, ') % _[0] for _ in tensors])))
  do_scope()
  for tensor in tensors:
    print_unload_buffer(printer, *tensor)

  for output_name in stencil.output_names:
    output_type = stencil.tensors[output_name].haoda_type
    println('if(var_%s_host_and_dev_are_null)' % output_name)
    do_scope('var_%s_host_and_dev_are_null' % output_name)
    output_str = [", 0, 0, 0"] * 4
    for dim in range(stencil.dim):
      stride = '1'
      if dim > 0:
        stride = '*'.join(
            ('%s_size_dim_%d' % (output_name, d)) for d in range(dim))
      output_str[dim] = (', var_{0}_min_{1}, {0}_size_dim_{1}, {2}'.format(
          output_name, dim, stride))
    println('bool %s = halide_rewrite_buffer(var_%s_buffer, %d%s%s%s%s);' %
            (new_var(), output_name, output_type.width_in_bytes,
             *output_str))
    println('(void)%s;' % last_var())
    un_scope()

  for input_name in stencil.input_names:
    input_type = stencil.tensors[input_name].haoda_type
    println('if(var_%s_host_and_dev_are_null)' % input_name)
    do_scope('var_%s_host_and_dev_are_null' % input_name)
    input_size = ['0'] * 4
    for dim in range(stencil.dim):
      println('int32_t %s = %s_size_dim_%d + %d;' %
              (new_var(), stencil.output_names[0], dim,
               core.get_stencil_dim(
                   core.get_overall_stencil_window(
                       stencil.tensors[stencil.input_names[0]],
                       stencil.tensors[stencil.output_names[0]]))[dim] - 1))
    input_str = [', 0, 0, 0'] * 4
    for dim in range(stencil.dim):
      stride = '1'
      if dim > 0:
        stride = '*'.join([last_var(x - stencil.dim) for x in range(dim)])
      input_str[dim] = (
          ', var_%s_min_%d, %s, %s' %
          (stencil.output_names[0], dim, last_var(dim - stencil.dim), stride))
    println('bool %s = halide_rewrite_buffer(var_%s_buffer, %d%s%s%s%s);' %
            (new_var(), input_name, input_type.width_in_bytes,
             *input_str))
    println('(void)%s;' % last_var())
    un_scope()

  println('bool %s = %s;' % (new_var(), ' || '.join(
      'var_%s_host_and_dev_are_null' % _
      for _ in stencil.output_names + stencil.input_names)))
  println('bool %s = !(%s);' % (new_var(), last_var(-2)))
  println('if(%s)' % last_var())
  do_scope('if(%s)' % last_var())

  for name in stencil.output_names + stencil.input_names + stencil.param_names:
    print_check_elem_size(printer, name, stencil.tensors[name].haoda_type)
  println()

  println('// allocate buffer for tiled input/output')
  for d in range(stencil.dim - 1):
    println('int32_t tile_num_dim_{d} = ({}_size_dim_{d}-STENCIL_DIM_{d}+1+'
            'TILE_SIZE_DIM_{d}-STENCIL_DIM_{d})/(TILE_SIZE_DIM_{d}-'
            'STENCIL_DIM_{d}+1);'.format(stencil.input_names[0], d=d))
  println()

  def for_banks():
    return printer.for_(
        'auto bank',
        '{{{}}}'.format(', '.join(map(str, range(soda_util.MAX_DRAM_BANK)))))

  println('// change #bank if there is a env var defined')
  println('unordered_map<string, array<bool, %d>> use_bank;' %
          soda_util.MAX_DRAM_BANK)
  println('unordered_map<string, int> num_bank;')
  println('unordered_map<string, array<int, %d>> bank_vec;' %
          soda_util.MAX_DRAM_BANK)
  for stmt in stencil.output_stmts + stencil.input_stmts + stencil.param_stmts:
    name = stmt.name
    println('use_bank["{}"] = {{{}}};'.format(
        name, ', '.join('true' if _ in stmt.dram else 'false'
                        for _ in range(soda_util.MAX_DRAM_BANK))))
    println('num_bank["{}"] = {{{}}};'.format(name, len(stmt.dram)))
    println('bank_vec["{}"] = {{{}}};'.format(name,
                                              ', '.join(map(str, stmt.dram))))
  for env_var in 'DRAM_IN', 'DRAM_OUT':
    with printer.if_('const char* env_var_char = getenv("{}")'.format(env_var)):
      println('const string env_var(env_var_char);')
      with printer.if_("env_var.find(':') != string::npos"):
        println('size_t pos = 0;')
        println('size_t next_pos;')
        with printer.do_while('next_pos != string::npos'):
          println("next_pos = env_var.find(',', pos);")
          println("const size_t colon_pos = env_var.find(':', pos);")
          println("const string name = env_var.substr(pos, colon_pos - pos);")
          println('pos = colon_pos + 1;')
          with printer.if_('use_bank.count(name)'):
            println('fill(use_bank[name].begin(), use_bank[name].end(), '
                    'false);')
            println('num_bank[name] = 0;')
            println('int idx = 0;')
            println('size_t dot_pos;')
            with printer.do_while('dot_pos != string::npos && '
                                  'dot_pos < next_pos'):
              println("dot_pos = env_var.find('.', pos);")
              println('int bank = stoi(env_var.substr(pos, dot_pos - pos));')
              println('use_bank[name][bank] = true;')
              println('++num_bank[name];')
              println('bank_vec[name][idx++] = bank;')
              println('pos = dot_pos + 1;')
            with printer.else_():
              println('fprintf(*error_report, '
                      '"WARNING: unknown name %s\\n", name.c_str());')
          println('pos = next_pos + 1;')
        with printer.else_():
          if env_var == 'DRAM_IN':
            var_names = stencil.input_names
          else:
            var_names = stencil.output_names
          with printer.for_('const string name',
                            '{%s}' % ', '.join(map('"{}"'.format, var_names))):
            println('fill(use_bank[name].begin(), use_bank[name].end(), '
                    'false);')
            println('num_bank[name] = 0;')
            println('int idx = 0;')
            println('size_t pos = 0;')
            println('size_t dot_pos;')
            with printer.do_while('dot_pos != string::npos'):
              println("dot_pos = env_var.find('.', pos);")
              println('int bank = stoi(env_var.substr(pos, dot_pos - pos));')
              println('use_bank[name][bank] = true;')
              println('++num_bank[name];')
              println('bank_vec[name][idx++] = bank;')
              println('pos = dot_pos + 1;')
    println()

  println('// align each linearized tile to multiples of BURST_WIDTH')
  println('int64_t tile_pixel_num = %s*%s_size_dim_%d;' %
          ('*'.join('TILE_SIZE_DIM_%d' % x for x in range(stencil.dim - 1)),
           stencil.input_names[0], stencil.dim - 1))
  println('int64_t tile_burst_num = '
          '(tile_pixel_num-1)/(BURST_WIDTH/%d*num_bank["%s"])+1;' %
          (stencil.input_stmts[0].haoda_type.width_in_bits,
           stencil.input_names[0]))
  println('int64_t tile_size_linearized_i = tile_burst_num*(BURST_WIDTH/'
          '{stmt.width_in_bits}*num_bank["{stmt.name}"]);'.format(
              stmt=stencil.input_stmts[0]))
  println('int64_t tile_size_linearized_o = tile_burst_num*(BURST_WIDTH/'
          '{stmt.width_in_bits}*num_bank["{stmt.name}"]);'.format(
              stmt=stencil.output_stmts[0]))
  println()

  println('// prepare for opencl')
  println('int err;')
  println()
  println('cl_platform_id platforms[16];')
  println('cl_platform_id platform_id = 0;')
  println('cl_uint platform_count;')
  println('cl_device_id device_id;')
  println('cl_context context;')
  println('cl_command_queue commands;')
  println('cl_program program;')
  println('cl_kernel kernel;')
  println()
  println('char cl_platform_vendor[1001];')
  println()

  for name in stencil.input_names + stencil.output_names:
    println('cl_mem var_%s_cl_bank[%d] = {};' % (name, soda_util.MAX_DRAM_BANK))
  println()

  for name in stencil.param_names:
    println('cl_mem var_%s_cl = nullptr;' % name)
  println()

  for stmt in stencil.input_stmts:
    println('uint64_t var_{stmt.name}_buf_size = sizeof({stmt.c_type})*({}*tile'
            '_size_linearized_i/num_bank["{stmt.name}"]+((STENCIL_DISTANCE-1)/('
            'BURST_WIDTH/{stmt.width_in_bits}*num_bank["{stmt.name}"])+1)*(BURS'
            'T_WIDTH/{stmt.width_in_bits}));'.format('*'.join(
                'tile_num_dim_%d' % _ for _ in range(stencil.dim - 1)),
                                                     stmt=stmt))
  for stmt in stencil.output_stmts:
    println('uint64_t var_{stmt.name}_buf_size = sizeof({stmt.c_type})*({}*tile'
            '_size_linearized_o/num_bank["{stmt.name}"]+((STENCIL_DISTANCE-1)/('
            'BURST_WIDTH/{stmt.width_in_bits}*num_bank["{stmt.name}"])+1)*(BURS'
            'T_WIDTH/{stmt.width_in_bits}));'.format('*'.join(
                'tile_num_dim_%d' % x for x in range(stencil.dim - 1)),
                                                     stmt=stmt))
  println()

  println('unsigned char *kernel_binary;')
  println('const char *device_name;')
  println('char target_device_name[64];')
  println('fprintf(*error_report, "INFO: Loading %s\\n", xclbin);')
  println('int64_t kernel_binary_size = load_xclbin2_to_memory(xclbin, (char **'
          ') &kernel_binary, (char**)&device_name);')
  println('if (strlen(device_name) == 0)')
  do_scope()
  println('device_name = getenv("XDEVICE");')
  un_scope()
  println('if(kernel_binary_size < 0)')
  do_scope()
  println('fprintf(*error_report, "ERROR: Failed to load kernel from xclbin: %s'
          '\\n", xclbin);')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println('for(int i = 0; i<64; ++i)')
  do_scope()
  println("target_device_name[i] = (device_name[i]==':'||device_name[i]=='.') ?"
          " '_' : device_name[i];")
  un_scope()
  println()

  println('err = clGetPlatformIDs(16, platforms, &platform_count);')
  println('if(err != CL_SUCCESS)')
  do_scope()
  println('fprintf(*error_report, "ERROR: Failed to find an OpenCL platform\\n"'
          ');')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println('fprintf(*error_report, "INFO: Found %d platforms\\n", platform_count'
          ');')
  println()

  println('int platform_found = 0;')
  println('for (unsigned iplat = 0; iplat<platform_count; iplat++)')
  do_scope()
  println('err = clGetPlatformInfo(platforms[iplat], CL_PLATFORM_VENDOR, 1000, '
          '(void *)cl_platform_vendor,nullptr);')
  println('if(err != CL_SUCCESS)')
  do_scope()
  println('fprintf(*error_report, "ERROR: clGetPlatformInfo(CL_PLATFORM_VENDOR)'
          ' failed\\n");')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println('if(strcmp(cl_platform_vendor, "Xilinx") == 0)')
  do_scope()
  println('fprintf(*error_report, "INFO: Selected platform %d from %s\\n", ipla'
          't, cl_platform_vendor);')
  println('platform_id = platforms[iplat];')
  println('platform_found = 1;')
  un_scope()
  un_scope()
  println('if(!platform_found)')
  do_scope()
  println('fprintf(*error_report, "ERROR: Platform Xilinx not found\\n");')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println()

  println('cl_device_id devices[16];')
  println('cl_uint device_count;')
  println('unsigned int device_found = 0;')
  println('char cl_device_name[1001];')
  println('err = clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_ACCELERATOR,')
  println('           16, devices, &device_count);')
  println('if(err != CL_SUCCESS)')
  do_scope()
  println('fprintf(*error_report, "ERROR: Failed to create a device group\\n")'
          ';')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println()

  println('for (unsigned int i=0; i<device_count; ++i)')
  do_scope()
  println('err = clGetDeviceInfo(devices[i], CL_DEVICE_NAME, 1024, cl_device_na'
          'me, 0);')
  println('if(err != CL_SUCCESS)')
  do_scope()
  println('fprintf(*error_report, "ERROR: Failed to get device name for device '
          '%d\\n", i);')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println('printf("INFO: Find device %s\\n", cl_device_name);')
  println('if(strcmp(cl_device_name, target_device_name) == 0 || strcmp(cl_devi'
          'ce_name, device_name) == 0)')
  do_scope()
  println('device_id = devices[i];')
  println('device_found = 1;')
  println('fprintf(*error_report, "INFO: Selected %s as the target device\\n",'
          ' device_name);')
  un_scope()
  un_scope()
  println()

  println('if(!device_found)')
  do_scope()
  println('fprintf(*error_report, "ERROR: Target device %s not found\\n", targe'
          't_device_name);')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println()

  println('err = clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_ACCELERATOR,')
  println('           1, &device_id, nullptr);')
  println('if(err != CL_SUCCESS)')
  do_scope()
  println('fprintf(*error_report, "ERROR: Failed to create a device group\\n")'
          ';')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println()

  println('context = clCreateContext(nullptr, 1, &device_id, nullptr, nullptr, '
          '&err);')
  println('if(!context)')
  do_scope()
  println('fprintf(*error_report, "ERROR: Failed to create a compute context\\n'
          '");')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println()

  println('commands = clCreateCommandQueue(context, device_id, 0, &err);')
  println('if(!commands)')
  do_scope()
  println('fprintf(*error_report, "ERROR: Failed to create a command commands %'
          'i\\n",err);')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println()

  println('int status;')
  println('size_t kernel_binary_sizes[1] = {static_cast<size_t>('
          'kernel_binary_size)};')
  println('program = clCreateProgramWithBinary(context, 1, &device_id, kernel_b'
          'inary_sizes,')
  println('                   (const unsigned char **) &kernel_binary, &status,'
          ' &err);')
  println('if((!program) || (err!=CL_SUCCESS))')
  do_scope()
  println('fprintf(*error_report, "ERROR: Failed to create compute program from'
          ' binary %d\\n", err);')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println('delete[] kernel_binary;')
  println()

  println('err = clBuildProgram(program, 0, nullptr, nullptr, nullptr, nullptr)'
          ';')
  println('if(err != CL_SUCCESS)')
  do_scope()
  println('size_t len;')
  println('char buffer[2048];')
  println('fprintf(*error_report, "ERROR: Failed to build program executable\\n'
          '");')
  println('clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, size'
          'of(buffer), buffer, &len);')
  println('fprintf(*error_report, "%s\\n", buffer);')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println()

  println('kernel = clCreateKernel(program, "%s_kernel", &err);' %
          stencil.app_name)
  println('if(!kernel || err != CL_SUCCESS)')
  do_scope()
  println('fprintf(*error_report, "ERROR: Failed to create compute kernel %d\\n'
          '", err);')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println()

  println(
      'const unsigned XCL_MEM_DDR_BANK[%d] = {%s};' %
      (soda_util.MAX_DRAM_BANK, ', '.join(
          'XCL_MEM_DDR_BANK%d' % _ for _ in range(soda_util.MAX_DRAM_BANK))))
  println('unordered_map<string, array<cl_mem_ext_ptr_t, %d>> mem_ext_bank;' %
          soda_util.MAX_DRAM_BANK)
  with printer.for_('const auto& use_bank_pair', 'use_bank'):
    with for_banks():
      with printer.if_('use_bank_pair.second[bank]'):
        println('mem_ext_bank[use_bank_pair.first][bank].flags = '
                'XCL_MEM_DDR_BANK[bank];')

  with for_banks():
    for name in stencil.input_names:
      with printer.if_('use_bank["{name}"][bank]'.format(name=name)):
        println('var_{name}_cl_bank[bank] = clCreateBuffer(context, '
                'CL_MEM_READ_ONLY|CL_MEM_EXT_PTR_XILINX, var_{name}_buf_size, '
                '&mem_ext_bank["{name}"][bank], nullptr);'.format(name=name))
        with printer.if_('var_{name}_cl_bank[bank] == nullptr'
                         ''.format(name=name)):
          println('fprintf(*error_report, '
                  '"ERROR: Failed to allocate device memory\\n");')
          println('exit(EXIT_FAILURE);')
    for name in stencil.output_names:
      with printer.if_('use_bank["{name}"][bank]'.format(name=name)):
        println('var_{name}_cl_bank[bank] = clCreateBuffer(context, '
                'CL_MEM_WRITE_ONLY|CL_MEM_EXT_PTR_XILINX, var_{name}_buf_size, '
                '&mem_ext_bank["{name}"][bank], nullptr);'.format(name=name))
        with printer.if_('var_{name}_cl_bank[bank] == nullptr'
                         ''.format(name=name)):
          println('fprintf(*error_report, '
                  '"ERROR: Failed to allocate device memory\\n");')
          println('exit(EXIT_FAILURE);')

  for param in stencil.param_stmts:
    println('var_%s_cl = clCreateBuffer(context, CL_MEM_READ_ONLY, '
            '%s*sizeof(%s), nullptr, nullptr);' %
            (param.name, '*'.join(map(str, param.size)), param.type))
    with printer.if_('var_%s_cl == nullptr' % param.name):
      println('fprintf(*error_report, '
              '"ERROR: Failed to allocate device memory\\n");')
      println('exit(EXIT_FAILURE);')
  println()

  println('cl_event write_events[%d];' %
          (soda_util.MAX_DRAM_BANK * len(stencil.input_stmts) +
           len(stencil.param_stmts)))
  println('cl_event* write_event_ptr = write_events;')
  for param in stencil.param_stmts:
    println('%s* var_%s_buf = (%s*)clEnqueueMapBuffer(commands, var_%s_cl, '
            'CL_FALSE, CL_MAP_WRITE, 0, %s*sizeof(%s), 0, nullptr, '
            'write_event_ptr++, &err);' %
            (param.c_type, param.name, param.c_type, param.name, '*'.join(
                map(str, param.size)), param.c_type))
  for stmt in stencil.input_stmts:
    println('{stmt.c_type}* var_{stmt.name}_buf_bank[{}] = {{}};'.format(
        soda_util.MAX_DRAM_BANK, stmt=stmt))
  for stmt in stencil.input_stmts:
    with for_banks():
      with printer.if_('use_bank["{name}"][bank]'.format(name=stmt.name)):
        println('var_{stmt.name}_buf_bank[bank] = '
                'reinterpret_cast<{stmt.c_type}*>(clEnqueueMapBuffer(commands,'
                ' var_{stmt.name}_cl_bank[bank], CL_FALSE, '
                'CL_MAP_WRITE, 0, var_{stmt.name}_buf_size, 0, nullptr, '
                'write_event_ptr++, &err));'.format(stmt=stmt))
  println('clWaitForEvents(write_event_ptr - write_events, write_events);')
  println()

  println('// tiling')
  for dim in range(stencil.dim - 2, -1, -1):
    println('for(int32_t tile_index_dim_{0} = 0; tile_index_dim_{0} < '
            'tile_num_dim_{0}; ++tile_index_dim_{0})'.format(dim))
    do_scope()
    println('int32_t actual_tile_size_dim_{0} = (tile_index_dim_{0}==tile_num_'
            'dim_{0}-1) ? {1}_size_dim_{0}-(TILE_SIZE_DIM_{0}-STENCIL_DIM_{0}+'
            '1)*tile_index_dim_{0} : TILE_SIZE_DIM_{0};'.format(
                dim, stencil.input_names[0]))

  println('#pragma omp parallel for', 0)
  println('for(int32_t {0} = 0; {0} < {1}_size_dim_{2}; ++{0})'.format(
      soda_util.COORDS_IN_TILE[stencil.dim - 1], stencil.input_names[0],
      stencil.dim - 1))
  do_scope()
  for dim in range(stencil.dim - 2, -1, -1):
    println('for(int32_t {0} = 0; {0} < actual_tile_size_dim_{1}; '
            '++{0})'.format(soda_util.COORDS_IN_TILE[dim], dim))
    do_scope()

  println('// (%s) is coordinates in tiled image' %
          ', '.join(soda_util.COORDS_TILED))
  println('// (%s) is coordinates in original image' %
          ', '.join(soda_util.COORDS_IN_ORIG))
  println('// (%s) is coordinates in a tile' %
          ', '.join(soda_util.COORDS_IN_TILE))
  offset_in_tile = '+'.join(
      '%c%s' % (soda_util.COORDS_IN_TILE[x], ''.join('*TILE_SIZE_DIM_%d' % xx
                                                     for xx in range(x)))
      for x in range(stencil.dim))
  println('int32_t burst_index = ({}) / (BURST_WIDTH / {stmt.width_in_bits} * '
          'num_bank["{stmt.name}"]);'.format(offset_in_tile,
                                             stmt=stencil.input_stmts[0]))
  println('int32_t burst_residue = ({}) % (BURST_WIDTH / {stmt.width_in_bits} '
          '* num_bank["{stmt.name}"]);'.format(offset_in_tile,
                                               stmt=stencil.input_stmts[0]))
  for dim in range(stencil.dim - 1):
    println('int32_t {0} = tile_index_dim_{1}*(TILE_SIZE_DIM_{1}-STENCIL_DIM_'
            '{1}+1)+{2};'.format(soda_util.COORDS_IN_ORIG[dim], dim,
                                 soda_util.COORDS_IN_TILE[dim]))
  println('int32_t %c = %c;' % (soda_util.COORDS_IN_ORIG[stencil.dim - 1],
                                soda_util.COORDS_IN_TILE[stencil.dim - 1]))
  println('int64_t tiled_offset = ({}) * tile_size_linearized_i + burst_index *'
          ' (BURST_WIDTH / {stmt.width_in_bits} * num_bank["{stmt.name}"]) + bu'
          'rst_residue;'.format('+'.join('%stile_index_dim_%d' %
                                         (''.join('tile_num_dim_%d*' % xx
                                                  for xx in range(x)), x)
                                         for x in range(stencil.dim - 1)),
                                stmt=stencil.input_stmts[0]))
  println('int64_t original_offset = %s;' %
          '+'.join('%c*var_%s_stride_%d' %
                   (soda_util.COORDS_IN_ORIG[x], stencil.input_names[0], x)
                   for x in range(stencil.dim)))
  for name in stencil.input_names:
    println('var_{name}_buf_bank[bank_vec["{name}"]'
            '[tiled_offset % num_bank["{name}"]]]'
            '[tiled_offset / num_bank["{name}"]] = '
            'var_{name}[original_offset];'.format(name=name))
  for dim in range(stencil.dim * 2 - 1):
    un_scope()
  println()

  for param in stencil.param_stmts:
    println('memcpy(var_{0}_buf, var_{0}, {1}*sizeof({2}));'.format(
        param.name, '*'.join(map(str, param.size)), param.c_type))
  if stencil.param_stmts:
    println()

  println('err = 0;')
  for idx, param in enumerate(stencil.param_stmts):
    println('err |= clEnqueueUnmapMemObject(commands, var_{0}_cl, var_{0}_buf, '
            '0, nullptr, write_events+{1});'.format(param.name, idx))
  for name in stencil.input_names:
    println('write_event_ptr = write_events;')
    with for_banks():
      with printer.if_('use_bank["{}"][bank]'.format(name)):
        println('err |= clEnqueueUnmapMemObject(commands, '
                'var_{name}_cl_bank[bank], var_{name}_buf_bank[bank], 0, '
                'nullptr, write_event_ptr++);'.format(name=name))
  println()

  println('if(err != CL_SUCCESS)')
  do_scope()
  println('fprintf(*error_report, "ERROR: Failed to write to input !\\n");')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println()

  println('err = 0;')
  for name in stencil.input_names + stencil.output_names:
    with for_banks():
      with printer.if_('use_bank["{name}"][bank]'.format(name=name)):
        println('fprintf(*error_report, "INFO: Using DRAM bank %d for {}\\n", '
                'bank);'.format(name))
  for d in range(stencil.dim - 1):
    println('fprintf(*error_report, "INFO: tile_num_dim_{0} = %d, '
            'TILE_SIZE_DIM_{0} = %d\\n", tile_num_dim_{0}, '
            'TILE_SIZE_DIM_{0});'.format(d))
  for name in stencil.input_names:
    println(
        'fprintf(*error_report, "INFO: %s\\n", %s);' %
        (', '.join(
            '%s_extent_%d = %%d' % (name, _) for _ in range(stencil.dim)),
         ', '.join('%s_size_dim_%d' % (name, _) for _ in range(stencil.dim))))
    println(
        'fprintf(*error_report, "INFO: %s\\n", %s);' %
        (', '.join('%s_min_%d = %%d' % (name, _) for _ in range(stencil.dim)),
         ', '.join('var_%s_min_%d' % (name, _) for _ in range(stencil.dim))))
  for name in stencil.output_names:
    println(
        'fprintf(*error_report, "INFO: %s\\n", %s);' %
        (', '.join(
            '%s_extent_%d = %%d' % (name, _) for _ in range(stencil.dim)),
         ', '.join('%s_size_dim_%d' % (name, _) for _ in range(stencil.dim))))
    println(
        'fprintf(*error_report, "INFO: %s\\n", %s);' %
        (', '.join('%s_min_%d = %%d' % (name, _) for _ in range(stencil.dim)),
         ', '.join('var_%s_min_%d' % (name, _) for _ in range(stencil.dim))))
  println()

  println('int kernel_arg = 0;')
  println('int64_t tile_data_num = ((int64_t({stmt.name}_size_dim_{dim}){} - 1)'
          ' / (BURST_WIDTH / {stmt.width_in_bits} * num_bank["{stmt.name}"]) + '
          '1) * BURST_WIDTH / {stmt.width_in_bits} * num_bank["{stmt.name}"] / '
          'UNROLL_FACTOR;'.format(''.join(
              ' * TILE_SIZE_DIM_%d' % _ for _ in range(stencil.dim - 1)),
                                  stmt=stencil.input_stmts[0],
                                  dim=stencil.dim - 1))
  println('int64_t coalesced_data_num = ((int64_t({stmt.name}_size_dim_{dim}){}'
          ' * {} + STENCIL_DISTANCE - 1) / (BURST_WIDTH / {stmt.width_in_bits} '
          '* num_bank["{stmt.name}"]) + 1);'.format(
              ''.join('*TILE_SIZE_DIM_%d' % _ for _ in range(stencil.dim - 1)),
              '*'.join('tile_num_dim_%d' % _ for _ in range(stencil.dim - 1)),
              stmt=stencil.input_stmts[0],
              dim=stencil.dim - 1))
  println('fprintf(*error_report, "INFO: tile_data_num = %ld, coalesced_data_nu'
          'm = %ld\\n", tile_data_num, coalesced_data_num);')
  println()

  for name in stencil.output_names + stencil.input_names:
    with for_banks():
      with printer.if_('use_bank["{name}"][bank]'.format(name=name)):
        println('err |= clSetKernelArg(kernel, kernel_arg++, '
                'sizeof(cl_mem), &var_%s_cl_bank[bank]);' % name)
  for name in stencil.param_names:
    println('err |= clSetKernelArg(kernel, kernel_arg++, sizeof(cl_mem), '
            '&var_%s_cl);' % name)
  println('err |= clSetKernelArg(kernel, kernel_arg++, sizeof(coalesced_data_nu'
          'm), &coalesced_data_num);')
  println('if(err != CL_SUCCESS)')
  do_scope()
  println('fprintf(*error_report, "ERROR: Failed to set kernel arguments %d'
          '\\n", err);')
  println('exit(EXIT_FAILURE);')
  un_scope()
  println()

  println('timespec execute_begin, execute_end;')
  println('cl_event execute_event;')
  println('err = clEnqueueTask(commands, kernel, '
          'write_event_ptr - write_events, write_events, &execute_event);')
  with printer.if_('getenv("XCL_EMULATION_MODE") == nullptr'):
    println('fprintf(*error_report, "INFO: FPGA warm up\\n");')
    println('clWaitForEvents(1, &execute_event);')
    println()
    println('clock_gettime(CLOCK_MONOTONIC_RAW, &execute_begin);')
    println('err = clEnqueueTask(commands, kernel, 0, nullptr, '
            '&execute_event);')
    with printer.if_('err'):
      println('fprintf(*error_report, "ERROR: Failed to execute kernel %d\\n", '
              'err);')
      println('  exit(EXIT_FAILURE);')
    println('clWaitForEvents(1, &execute_event);')
    println('clock_gettime(CLOCK_MONOTONIC_RAW, &execute_end);')
    println()
    println('double elapsed_time = 0.;')
    println('elapsed_time = (double(execute_end.tv_sec-execute_begin.tv_sec)+'
            '(execute_end.tv_nsec-execute_begin.tv_nsec)/1e9)*1e6;')
    println('printf("Kernel execution time: %lf us\\n", elapsed_time);')
    println('printf("Kernel throughput:   %%lf pixel/ns\\n", '
            '%s / elapsed_time / 1e3);' % '*'.join('%s_size_dim_%d' %
                                                   (stencil.input_names[0], _)
                                                   for _ in range(stencil.dim)))
    with printer.else_():
      println('clWaitForEvents(1, &execute_event);')
      println('fprintf(*error_report, "INFO: Emulation mode\\n");')
  println()

  println('cl_event read_events[%d];' %
          (soda_util.MAX_DRAM_BANK * len(stencil.output_stmts)))
  println('cl_event* read_event_ptr = read_events;')
  for stmt in stencil.output_stmts:
    println('{stmt.c_type}* var_{stmt.name}_buf_bank[{}] = {{}};'.format(
        soda_util.MAX_DRAM_BANK, stmt=stmt))
  for stmt in stencil.output_stmts:
    with for_banks():
      with printer.if_('use_bank["{name}"][bank]'.format(name=stmt.name)):
        println('var_{stmt.name}_buf_bank[bank] = '
                'reinterpret_cast<{stmt.c_type}*>(clEnqueueMapBuffer('
                'commands, var_{stmt.name}_cl_bank[bank], CL_FALSE, '
                'CL_MAP_READ, 0, var_{stmt.name}_buf_size, 0, nullptr, '
                'read_event_ptr++, &err));'.format(stmt=stmt))
  println('clWaitForEvents(read_event_ptr - read_events, read_events);')
  println()

  for dim in range(stencil.dim - 2, -1, -1):
    println('for(int32_t tile_index_dim_{0} = 0; tile_index_dim_{0} < '
            'tile_num_dim_{0}; ++tile_index_dim_{0})'.format(dim))
    do_scope()
    println('int32_t actual_tile_size_dim_{0} = (tile_index_dim_{0}==tile_num_d'
            'im_{0}-1) ? {1}_size_dim_{0}-(TILE_SIZE_DIM_{0}-STENCIL_DIM_{0}+1)'
            '*tile_index_dim_{0} : TILE_SIZE_DIM_{0};'.format(
                dim, stencil.input_names[0]))

  if stencil.preserve_border:
    overall_stencil_window = util.get_overall_stencil_window(
        stencil.output.parent.preserve_border_from(), stencil.output)
    overall_stencil_offset = util.get_stencil_window_offset(
        overall_stencil_window)
    overall_stencil_dim = util.get_stencil_dim(overall_stencil_window)
    println('for(int32_t {var} = 0; {var} < {0}_size_dim_{1}; ++{var})'.format(
        stencil.input.name,
        stencil.dim - 1,
        var=soda_util.COORDS_IN_TILE[stencil.dim - 1]))
    do_scope()
    for dim in range(stencil.dim - 2, -1, -1):
      println('for(int32_t {var} = tile_index_dim_{dim}==0 ? 0 : {0}; {var} < a'
              'ctual_tile_size_dim_{dim}-(tile_index_dim_{dim}==tile_num_dim_{d'
              'im}-1 ? 0 : {1}); ++{var})'.format(
                  overall_stencil_offset[dim],
                  overall_stencil_dim[dim] - 1 - overall_stencil_offset[dim],
                  dim=dim,
                  var=soda_util.COORDS_IN_TILE[dim]))
      do_scope()
  else:
    overall_stencil_window = core.get_overall_stencil_window(
        stencil.tensors[stencil.input_names[0]],
        stencil.tensors[stencil.output_names[0]])
    overall_stencil_offset = core.get_stencil_window_offset(
        overall_stencil_window)
    overall_stencil_dim = core.get_stencil_dim(overall_stencil_window)
    println('#pragma omp parallel for', 0)
    println('for(int32_t {var} = {}; {var} < {}_size_dim_{}-{}; ++{var})'
            ''.format(overall_stencil_offset[stencil.dim - 1],
                      stencil.output_names[0],
                      stencil.dim - 1,
                      overall_stencil_dim[stencil.dim - 1] - 1 -
                      overall_stencil_offset[stencil.dim - 1],
                      var=soda_util.COORDS_IN_TILE[stencil.dim - 1]))
    do_scope()
    for dim in range(stencil.dim - 2, -1, -1):
      println('for(int32_t {var} = {}; {var} < actual_tile_size_dim_{}-{}; '
              '++{var})'.format(overall_stencil_offset[dim],
                                dim,
                                overall_stencil_dim[dim] - 1 -
                                overall_stencil_offset[dim],
                                var=soda_util.COORDS_IN_TILE[dim]))
      do_scope()

  println('// (%s) is coordinates in tiled image' %
          ', '.join(soda_util.COORDS_TILED))
  println('// (%s) is coordinates in original image' %
          ', '.join(soda_util.COORDS_IN_ORIG))
  println('// (%s) is coordinates in a tile' %
          ', '.join(soda_util.COORDS_IN_TILE))
  offset_in_tile = '+'.join(
      '%c%s' % (soda_util.COORDS_IN_TILE[x], ''.join('*TILE_SIZE_DIM_%d' % xx
                                                     for xx in range(x)))
      for x in range(stencil.dim))
  for dim in range(stencil.dim - 1):
    println('int32_t {0} = tile_index_dim_{1}*(TILE_SIZE_DIM_{1}-STENCIL_DIM_{1'
            '}+1)+{2};'.format(soda_util.COORDS_IN_ORIG[dim], dim,
                               soda_util.COORDS_IN_TILE[dim]))
  println('int32_t %c = %c;' % (soda_util.COORDS_IN_ORIG[stencil.dim - 1],
                                soda_util.COORDS_IN_TILE[stencil.dim - 1]))
  println('int64_t original_offset = %s;' %
          '+'.join('%c*var_%s_stride_%d' %
                   (soda_util.COORDS_IN_ORIG[x], stencil.output_names[0], x)
                   for x in range(stencil.dim)))
  for stmt in stencil.output_stmts:
    overall_stencil_window = core.get_overall_stencil_window(
        map(stencil.tensors.get, stencil.input_names),
        stencil.tensors[stmt.name])
    overall_stencil_distance = core.get_stencil_distance(
        overall_stencil_window, stencil.tile_size)
    stencil_offset = overall_stencil_distance - soda_util.serialize(
        core.get_stencil_window_offset(overall_stencil_window),
        stencil.tile_size)
    println('int32_t burst_index_{stmt.name} = ({} + {}) / (BURST_WIDTH / {stmt'
            '.width_in_bits} * num_bank["{stmt.name}"]);'.format(offset_in_tile,
                                                                 stencil_offset,
                                                                 stmt=stmt))
    println('int32_t burst_residue_{stmt.name} = ({} + {}) % (BURST_WIDTH / {s'
            'tmt.width_in_bits} * num_bank["{stmt.name}"]);'.format(
                offset_in_tile, stencil_offset, stmt=stmt))
    println('int64_t tiled_offset_{stmt.name} = ({}) * tile_size_linearized_o +'
            ' burst_index_{stmt.name} * (BURST_WIDTH / {stmt.width_in_bits} * n'
            'um_bank["{stmt.name}"]) + burst_residue_{stmt.name};'.format(
                '+'.join('%stile_index_dim_%d' %
                         (''.join('tile_num_dim_%d*' % xx
                                  for xx in range(x)), x)
                         for x in range(stencil.dim - 1)),
                stmt=stmt))
    println(
        'var_{name}[original_offset] = '
        'var_{name}_buf_bank[bank_vec["{name}"]'
        '[tiled_offset_{name} % num_bank["{name}"]]]'
        '[tiled_offset_{name} / num_bank["{name}"]];'.format(name=stmt.name))
  for dim in range(stencil.dim * 2 - 1):
    un_scope()

  for name in stencil.output_names:
    println('read_event_ptr = read_events;')
    with for_banks():
      with printer.if_('use_bank["{name}"][bank]'.format(name=name)):
        println('clEnqueueUnmapMemObject(commands, var_{name}_cl_bank[bank], '
                'var_{name}_buf_bank[bank], 0, nullptr, read_event_ptr++);'
                ''.format(name=name))
    println('clWaitForEvents(read_event_ptr - read_events, read_events);')
  println()

  for name in stencil.input_names + stencil.output_names:
    with for_banks():
      with printer.if_('use_bank["{name}"][bank]'.format(name=name)):
        println('clReleaseMemObject(var_%s_cl_bank[bank]);' % name)
  for name in stencil.param_names:
    println('clReleaseMemObject(var_%s_cl);' % name)
  println()

  println('clReleaseProgram(program);')
  println('clReleaseKernel(kernel);')
  println('clReleaseCommandQueue(commands);')
  println('clReleaseContext(context);')

  un_scope()
  println('return 0;')
  un_scope()
  println()


def print_entrance(printer, stencil):
  println = printer.println
  tensors = [(stmt.name, stmt.c_type) for stmt in stencil.input_stmts +
             stencil.output_stmts + stencil.param_stmts]
  println('int %s(%sconst char* xclbin) HALIDE_FUNCTION_ATTRS' %
          (stencil.app_name, ''.join(
              'buffer_t *var_%s_buffer, ' % x[0] for x in tensors)))
  printer.do_scope()
  for tensor in tensors:
    print_unload_buffer(printer, *tensor)
  println('return %s_wrapped(%sxclbin);' % (stencil.app_name, ''.join(
      ('var_%s_buffer, ') % x[0] for x in tensors)))
  printer.un_scope()
  println()


def print_unload_buffer(printer, buffer_name, buffer_type):
  println = printer.println
  println('%s *var_%s = (%s *)(var_%s_buffer->host);' %
          (buffer_type, buffer_name, buffer_type, buffer_name))
  println('(void)var_%s;' % buffer_name)
  println('const bool var_{0}_host_and_dev_are_null = (var_{0}_buffer->host == '
          'nullptr) && (var_{0}_buffer->dev == 0);'.format(buffer_name))
  println('(void)var_%s_host_and_dev_are_null;' % buffer_name)
  for item in ['min', 'extent', 'stride']:
    for i in range(4):
      if item == 'extent':
        println('int32_t %s_size_dim_%d = var_%s_buffer->%s[%d];' %
                (buffer_name, i, buffer_name, item, i))
        println('(void)%s_size_dim_%d;' % (buffer_name, i))
      else:
        println('int32_t var_%s_%s_%d = var_%s_buffer->%s[%d];' %
                (buffer_name, item, i, buffer_name, item, i))
        println('(void)var_%s_%s_%d;' % (buffer_name, item, i))
  println('int32_t var_{0}_elem_size = var_{0}_buffer->elem_size;'.format(
      buffer_name))
  println('(void)var_%s_elem_size;' % buffer_name)


def print_check_elem_size(printer, buffer_name, buffer_type):
  println = printer.println
  new_var = printer.new_var
  last_var = printer.last_var
  println('bool %s = var_%s_elem_size == %d;' %
          (new_var(), buffer_name, buffer_type.width_in_bytes))
  println('if(!%s)' % last_var())
  printer.do_scope()
  println('int32_t %s = halide_error_bad_elem_size(nullptr, "Buffer %s", "%s",'
          ' var_%s_elem_size, %d);' %
          (new_var(), buffer_name, buffer_type, buffer_name,
           buffer_type.width_in_bytes))
  println('return %s;' % last_var())
  printer.un_scope()


def print_test(printer, stencil):
  println = printer.println
  do_scope = printer.do_scope
  un_scope = printer.un_scope

  input_dim = stencil.dim
  output_dim = stencil.dim

  println('int %s_test(const char* xclbin, const int dims[4])' %
          stencil.app_name)
  do_scope()
  for name in tuple(stencil.tensors) + stencil.param_names:
    println('buffer_t %s;' % name)
  for name in tuple(stencil.tensors) + stencil.param_names:
    println('memset(&%s, 0, sizeof(buffer_t));' % name)
  println()

  for tensor in stencil.tensors.values():
    println('%s* %s_img = new %s[%s]();' %
            (tensor.c_type, tensor.name, tensor.c_type, '*'.join(
                'dims[%d]' % x for x in range(stencil.dim))))
  for param in stencil.param_stmts:
    println('%s %s_img%s;' %
            (param.c_type, param.name,
             functools.reduce(operator.add, ['[%s]' % x for x in param.size])))
  println()

  for tensor in stencil.tensors.values():
    for d in range(stencil.dim):
      println('%s.extent[%d] = dims[%d];' % (tensor.name, d, d))
    println('%s.stride[0] = 1;' % tensor.name)
    for d in range(1, stencil.dim):
      println('%s.stride[%d] = %s;' %
              (tensor.name, d, '*'.join(['dims[%d]' % x for x in range(d)])))
    println('%s.elem_size = sizeof(%s);' % (tensor.name, tensor.c_type))
    println('{0}.host = (uint8_t*){0}_img;'.format(tensor.name))
    println()

  for param in stencil.param_stmts:
    for d, size in enumerate(param.size):
      println('%s.extent[%d] = %d;' % (param.name, d, size))
    println('%s.stride[0] = 1;' % param.name)
    for d in range(1, len(param.size)):
      println('%s.stride[%d] = %s;' %
              (param.name, d, '*'.join([str(x) for x in param.size[:d]])))
    println('%s.elem_size = sizeof(%s);' % (param.name, param.type))
    println('%s.host = (uint8_t*)%s_img;' % (param.name, param.name))
    println()

  if any(x.is_float for x in  stencil.input_types):
    println('std::default_random_engine generator;')
    println('std::uniform_real_distribution<double> distribution(0.0, 1.0);')
    println()

  for name in stencil.input_names:
    println('// initialization can be parallelized with -fopenmp')
    println('#pragma omp parallel for', 0)
    for d in range(0, stencil.dim):
      dim = stencil.dim - d - 1
      println('for(int32_t {var} = 0; {var}<dims[{}]; ++{var})'.format(
          dim, var=soda_util.COORDS_IN_ORIG[dim]))
      do_scope()
    init_val = '+'.join(soda_util.COORDS_IN_ORIG[0:input_dim])
    if stencil.symbol_table[name].is_float:
      init_val = 'distribution(generator)'
    println('%s_img[%s] = %s;' %
            (name, '+'.join('%c*%s.stride[%d]' %
                            (soda_util.COORDS_IN_ORIG[d], name, d)
                            for d in range(input_dim)), init_val))
    for d in range(0, input_dim):
      un_scope()
    println()

  for param in stencil.param_stmts:
    println('#pragma omp parallel for', 0)
    for dim, size in enumerate(param.size):
      println('for(int32_t {var} = 0; {var}<{}; ++{var})'.format(
          size, var=soda_util.COORDS_IN_ORIG[dim]))
      do_scope()
    println('%s_img%s = %s;' %
            (param.name,
             sum(('[%c]' % (soda_util.COORDS_IN_ORIG[d])
                  for d in range(len(param.size))), ''), '+'.join(
                      soda_util.COORDS_IN_ORIG[0:len(param.size)])))
    for d in param.size:
      un_scope()
    println()

  println('{}({}xclbin);'.format(
      stencil.app_name, ''.join(
          map('&{}, '.format, stencil.input_names + stencil.output_names +
              stencil.param_names))))
  println()

  println('int error_count = 0;')
  println()

  for tensor in stencil.chronological_tensors:
    if tensor.is_input():
      continue
    logger.debug('emit code to produce %s_img' % tensor.name)
    println('// produce %s, can be parallelized with -fopenmp' % tensor.name)
    println('#pragma omp parallel for', 0)
    stencil_window = core.get_overall_stencil_window(
        tuple(map(stencil.tensors.get, stencil.input_names)), tensor)
    stencil_dim = core.get_stencil_dim(stencil_window)
    output_idx = core.get_stencil_window_offset(stencil_window)
    for d in range(0, stencil.dim):
      dim = stencil.dim - d - 1
      println('for(int32_t {var} = {}; {var}<dims[{}]-{}; ++{var})'.format(
          output_idx[dim],
          dim,
          stencil_dim[dim] - output_idx[dim] - 1,
          var=soda_util.COORDS_IN_ORIG[dim]))
      do_scope()

    def mutate_load_for_host(obj, args):
      if isinstance(obj, ir.Ref):
        if obj.name in stencil.param_names:
          return ir.make_var(
              '%s_img%s' % (obj.name, ''.join('[%d]' % _ for _ in obj.idx)),
              haoda_type=obj.haoda_type,
          )
        return ir.make_var(
            '%s_img[%s]' %
            (obj.name, '+'.join('(%c%+d)*%s.stride[%d]' %
                                (soda_util.COORDS_IN_ORIG[d],
                                 obj.idx[d] - tensor.st_ref.idx[d], obj.name, d)
                                for d in range(stencil.dim))),
            haoda_type=obj.haoda_type,
        )
      return obj

    def mutate_store_for_host(obj, args):
      if isinstance(obj, ir.Ref):
        if obj.name in stencil.output_names:
          return ir.make_var(
              '%s result_%s' % (obj.c_type, obj.name),
              haoda_type=obj.haoda_type,
          )
        return ir.make_var(
            '%s_img[%s]' %
            (obj.name, '+'.join('%c*%s.stride[%d]' %
                                (soda_util.COORDS_IN_ORIG[d], obj.name, d)
                                for d in range(stencil.dim))),
            haoda_type=obj.haoda_type,
        )
      return obj

    for let in tensor.lets:
      println('// let {} {} = {}'.format(let.haoda_type, let.name, let.expr))
      println('const {} {} = {};'.format(
          let.c_type, let.name,
          let.expr.visit(mutate_load_for_host).c_expr))
    println('// {} = {}'.format(tensor.st_ref, tensor.expr))
    println('{} = {};'.format(tensor.st_ref.visit(mutate_store_for_host),
                              tensor.expr.visit(mutate_load_for_host).c_expr))
    if tensor.is_output():
      run_result = '%s_img[%s]' % (tensor.name, '+'.join(
          '%c*%s.stride[%d]' % (soda_util.COORDS_IN_ORIG[d], tensor.name, d)
          for d in range(stencil.dim)))
      println('%s val_fpga = %s;' % (tensor.c_type, run_result))
      println('%s val_cpu = result_%s;' % (tensor.c_type, tensor.name))
      if tensor.haoda_type.is_float:
        println('double threshold = 0.00001;')
        println('if(nullptr!=getenv("THRESHOLD"))')
        do_scope()
        println('threshold = atof(getenv("THRESHOLD"));')
        un_scope()
        println('threshold *= threshold;')
        println('if(double(val_fpga-val_cpu)*double(val_fpga-val_cpu) > '
                'threshold && '
                'double(val_fpga-val_cpu)*double(val_fpga-val_cpu)/(double'
                '(val_cpu)*double(val_cpu)) > threshold)')
        do_scope()
        params = (', '.join(['%d'] * stencil.dim),
                  ', '.join(soda_util.COORDS_IN_ORIG[:stencil.dim]))
        println('fprintf(*error_report, "%%lf != %%lf @(%s)\\n", double'
                '(val_fpga), double(val_cpu), %s);' % params)
      else:
        println('if(val_fpga!=val_cpu)')
        do_scope()
        params = (', '.join(['%d'] * stencil.dim),
                  ', '.join(soda_util.COORDS_IN_ORIG[:stencil.dim]))
        println('fprintf(*error_report, "%%ld != %%ld @(%s)\\n", int64_t'
                '(val_fpga), int64_t(val_cpu), %s);' % params)
      println('++error_count;')
      un_scope()
    for d in range(0, stencil.dim):
      un_scope()
    println()

  println('if(error_count==0)')
  do_scope()
  println('fprintf(*error_report, "INFO: PASS!\\n");')
  un_scope()
  println('else')
  do_scope()
  println('fprintf(*error_report, "INFO: FAIL!\\n");')
  un_scope()
  println()

  for var in (stencil.input_names + stencil.local_names + stencil.output_names +
              stencil.param_names):
    println('delete[] %s_img;' % var)
  println()

  println('return error_count;')
  un_scope()


def print_code(stencil, host_file):
  logger.info('generate host source code as %s' % host_file.name)
  printer = util.CppPrinter(host_file)
  println = printer.println
  do_scope = printer.do_scope
  un_scope = printer.un_scope
  print_define = lambda key, value: util.print_define(printer, key, value)

  print_header(printer)
  println('#include "%s.h"' % stencil.app_name)
  println()

  print_define('BURST_WIDTH', stencil.burst_width)
  for i in range(len(stencil.tile_size) - 1):
    util.print_define(printer, 'TILE_SIZE_DIM_%d' % i, stencil.tile_size[i])
  util.print_define(printer, 'UNROLL_FACTOR', stencil.unroll_factor)

  if stencil.preserve_border:
    overall_stencil_window = core.get_overall_stencil_window(
        stencil.output.parent.preserve_border_from(), stencil.output)
  else:
    overall_stencil_window = core.get_overall_stencil_window(
        map(stencil.tensors.get, stencil.input_names),
        stencil.tensors[stencil.output_names[0]])

  overall_stencil_distance = core.get_stencil_distance(overall_stencil_window,
                                                       stencil.tile_size)

  for i, dim in enumerate(core.get_stencil_dim(overall_stencil_window)):
    print_define('STENCIL_DIM_%d' % i, dim)
  stencil_offset = overall_stencil_distance - soda_util.serialize(
      core.get_stencil_window_offset(overall_stencil_window), stencil.tile_size)
  if stencil.preserve_border:
    stencil_offset *= stencil.iterate

  overall_stencil_distance = max(overall_stencil_distance, stencil_offset)

  print_define('STENCIL_DISTANCE', overall_stencil_distance)
  println()

  print_load_xclbin2(printer)
  print_halide_rewrite_buffer(printer)
  print_halide_error_codes(printer)
  print_halide_error_report(printer)
  print_wrapped(printer, stencil)
  print_entrance(printer, stencil)
  print_test(printer, stencil)



from typing import (
  Dict,
  Iterable,
  List,
  Mapping,
  Optional,
  Tuple,
  Union,
)

import collections

from haoda import ir
from soda import tensor

def get_load_tuple(obj: Union[ir.Node, tensor.Tensor]) -> Tuple[ir.Ref, ...]:
  """Get all load references as a tuple.

  Args:
    obj: A ir.Node object or a tensor.Tensor object.

  Returns:
    A tuple of all the load references.

  Raises:
    TypeError: If obj is not an IR node or a tensor.Tensor.
  """
  def visitor(obj: ir.Node, loads: List[ir.Ref]) -> ir.Node:
    if isinstance(obj, ir.Ref):
      loads.append(obj)
    return obj
  loads = []  # type: List[ir.Ref]
  if isinstance(obj, ir.Node):
    obj.visit(visitor, loads)
  elif isinstance(obj, tensor.Tensor):
    obj.visit_loads(visitor, loads)
  else:
    raise TypeError('argument is not an IR node or a tensor.Tensor')
  return tuple(loads)

def get_load_set(obj: Union[ir.Node, tensor.Tensor]) -> Tuple[ir.Ref, ...]:
  """Get all unique load references as a tuple.

  Args:
    obj: A haoda.ir.Node object or a tensor.Tensor object.

  Returns:
    A tuple of all unique loads.

  Raises:
    TypeError: If obj is not an IR node or a tensor.Tensor.
  """
  def visitor(obj: ir.Node, loads: Dict[ir.Ref, None]) -> ir.Node:
    if isinstance(obj, ir.Ref):
      loads[obj] = None
    return obj
  loads = collections.OrderedDict()   # type: Dict[ir.Ref, None]
  if isinstance(obj, ir.Node):
    obj.visit(visitor, loads)
  elif isinstance(obj, tensor.Tensor):
    obj.visit_loads(visitor, loads)
  else:
    raise TypeError('argument is not an IR node or a tensor.Tensor')
  return tuple(loads)

def get_load_dict(obj: Union[ir.Node, tensor.Tensor]) -> Dict[ir.Ref,
                                                            List[ir.Ref]]:
  """Get all load references as a dict mapping names to lists of loads.

  Args:
    obj: A haoda.ir.Node object or a tensor.Tensor object.

  Returns:
    A dict mapping accessed ir.Ref names to the corresponding lists of loads.

  Raises:
    TypeError: If obj is not an IR node or a tensor.Tensor.
  """
  def visitor(obj: ir.Node, loads: Dict[ir.Ref, List[ir.Ref]]) -> ir.Node:
    if isinstance(obj, ir.Ref):
      loads.setdefault(obj.name, []).append(obj)  # type: ignore
    return obj
  loads = collections.OrderedDict()   # type: Dict[ir.Ref, List[ir.Ref]]
  if isinstance(obj, tensor.Tensor):
    obj.visit_loads(visitor, loads)
  elif isinstance(obj, ir.Node):
    obj.visit(visitor, loads)
  else:
    raise TypeError('argument is not an IR node or a tensor.Tensor')
  return loads

def get_normalize_index(
    obj: Union[ir.Node, Iterable[ir.Node]],
    references: Optional[Mapping[str, Tuple[int, ...]]] = None
    ) -> Tuple[int, ...]:
  """Get the normalize index that will make the least access index 0.

  Args:
    obj: A node or an iterable of nodes.
  Returns:
    Normalize index as a tuple.
  Raises:
    TypeError: If argument is not an ir.Node or an iterable of ir.Nodes.
  """
  if not isinstance(obj, (collections.Iterable, ir.Node)):
    raise TypeError('argument is not an ir.Node or an iterable of ir.Nodes')
  if isinstance(obj, ir.Node):
    obj = (obj,)
  try:
    def get_idx(load: ir.Ref) -> Tuple[int, ...]:
      if references is None:
        return load.idx
      ref = references.get(getattr(load, 'name'))
      if ref is None:
        return load.idx
      return tuple(x - y for x, y in zip(load.idx, ref))
    return get_idx(min(sum(map(get_load_tuple, obj), ()),
                       key=lambda load: tuple(reversed(get_idx(load)))))
  except ValueError as e:
    if str(e) == 'min() arg is an empty sequence':
      return ()
    raise e



#!/usr/bin/python3
import collections
import json
import sys

class XilinxHLSReport():
  def __init__(self, rpt_file):
    def get_resource_dict(resources, usages):
      usage = collections.OrderedDict()
      for resource, used in zip(resources, usages):
        usage[resource] = int(used)
      return usage

    for line in rpt_file:
      if line.startswith('* Project:'):
        self.app = line.split(':')[1].split(' ')[-1][:-8]
      elif line == '== Utilization Estimates\n':
        self.resources_used = collections.OrderedDict()
        self.resources_available = collections.OrderedDict()
        self.resources_instances = collections.OrderedDict()
      elif (hasattr(self, 'resources_used') and
          not self.resources_used and 'Name' in line):
        for resource in line.split('|')[2:-1]:
          resource = resource.strip()
          self.resources_used[resource] = None
          self.resources_available[resource] = None
          self.resources_instances[resource] = None
        self.resources_used[None] = None
        self.resources_available[None] = None
        self.resources_instances[None] = None
      elif (hasattr(self, 'resources_used') and
          None in self.resources_used and '|Total' in line):
        for resource, used in zip(self.resources_used,
                      line.split('|')[2:-1]):
          self.resources_used[resource] = int(used)
        del self.resources_used[None]
      elif (hasattr(self, 'resources_available') and
          None in self.resources_available and '|Available' in line):
        for resource, avail in zip(self.resources_available,
                       line.split('|')[2:-1]):
          self.resources_available[resource] = int(avail)
        del self.resources_available[None]
      elif (hasattr(self, 'resources_instances') and
          None in self.resources_instances and '|Instance' in line):
        for resource, avail in zip(self.resources_instances,
                       line.split('|')[2:-1]):
          self.resources_instances[resource] = int(avail)
        del self.resources_instances[None]
      elif line.startswith('  * Instance:'):
        self.instances = collections.OrderedDict()
        self.instances[None] = None
      elif (hasattr(self, 'instances') and
          None in self.instances and '|compute_' in line):
        pe_id = line.split('|')[2].split('_')[-2]
        if pe_id[-1] == 'u':
          pe_id = int(pe_id[:-1])
        else:
          pe_id = int(pe_id)
        func_name = '_'.join(line.split('|')[2].split('_')[1:-2])
        compute = self.instances.setdefault('compute',
                                            collections.OrderedDict())
        compute = compute.setdefault(func_name, collections.OrderedDict())
        compute[pe_id] = get_resource_dict(self.resources_used,
                           line.split('|')[3:-1])
      elif (hasattr(self, 'instances') and
          None in self.instances and '|forward_' in line):
        data_type = line.split('|')[2].split('_')[2]
        fifo_depth = line.split('|')[2].split('_')[3]
        if fifo_depth[-1] == 'u':
          fifo_depth = int(fifo_depth[:-1])
        else:
          fifo_depth = int(fifo_depth)
        forward = self.instances.setdefault('forward',
                          collections.OrderedDict())
        forward = forward.setdefault(
          int(line.split('|')[2].split('_')[1]), collections.OrderedDict())
        forward = forward.setdefault(data_type, collections.OrderedDict())
        forward = forward.setdefault(fifo_depth, [])
        forward.append(get_resource_dict(self.resources_used,
                         line.split('|')[3:-1]))
      elif (hasattr(self, 'instances') and
          None in self.instances and '|'+self.app+'_kernel' in line):
        if line.split('|')[2].strip().endswith('_m_axi'):
          m_axi = self.instances.setdefault('m_axi', collections.OrderedDict())
          m_axi[line.split('|')[2].split('_')[2]] = get_resource_dict(
            self.resources_used, line.split('|')[3:-1])
        elif line.split('|')[2].strip().endswith('_control_s_axi'):
          self.instances['control_s_axi'] = get_resource_dict(
            self.resources_used, line.split('|')[3:-1])
        elif self.app+'_kernel_entry' in line.split('|')[2]:
          self.instances['entry'] = get_resource_dict(
            self.resources_used, line.split('|')[3:-1])
      elif (hasattr(self, 'instances') and
          None in self.instances and '|Block_proc' in line):
        self.instances['Block_proc'] = get_resource_dict(
          self.resources_used, line.split('|')[3:-1])
      elif (hasattr(self, 'instances') and
          None in self.instances and '|load' in line):
        self.instances.setdefault('load', []).append(
            get_resource_dict(self.resources_used,
                      line.split('|')[3:-1]))
      elif (hasattr(self, 'instances') and
          None in self.instances and '|store' in line):
        self.instances.setdefault('store', []).append(
            get_resource_dict(self.resources_used,
                      line.split('|')[3:-1]))
      elif (hasattr(self, 'instances') and
          None in self.instances and '|pack_' in line):
        pack = self.instances.setdefault('pack', collections.OrderedDict())
        pack = pack.setdefault(0, collections.OrderedDict())
        pack = pack.setdefault('float', [])
        pack.append(get_resource_dict(self.resources_used,
                        line.split('|')[3:-1]))
      elif (hasattr(self, 'instances') and
          None in self.instances and '|unpack_' in line):
        unpack = self.instances.setdefault('unpack', collections.OrderedDict())
        unpack = unpack.setdefault(0, collections.OrderedDict())
        unpack = unpack.setdefault('float', [])
        unpack.append(get_resource_dict(self.resources_used,
                        line.split('|')[3:-1]))
      elif (hasattr(self, 'instances') and
          None in self.instances and '|Total' in line):
        del self.instances[None]
    dram_bank_i = sum(1 for bus in self.instances['m_axi']
              if bus.endswith('i'))
    dram_bank_o = sum(1 for bus in self.instances['m_axi']
              if bus.endswith('o'))
    unroll_factor = len(next(iter(self.instances['compute'])))
    pack = self.instances['pack']
    unpack = self.instances['unpack']
    pack[unroll_factor//dram_bank_o] = pack.pop(0)
    unpack[unroll_factor//dram_bank_i] = unpack.pop(0)

  def __str__(self):
    return json.dumps(self.__dict__, indent=2, sort_keys=True)

  def check(self):
    resources = list(self.resources_instances)
    resources_used = collections.OrderedDict()
    for resource in resources:
      resources_used[resource] = 0
    for module in ('Block_proc', 'entry', 'control_s_axi'):
      if module in self.instances:
        accum_res(resources_used,
                   self.instances[module])
    for module in ('m_axi',):
      accum_res_iter(resources_used,
                       self.instances[module].values())
    for module in ('load', 'store'):
      accum_res_iter(resources_used,
                       self.instances[module])
    for module in ('unpack', 'pack'):
      for degree in self.instances[module].values():
        for data_type in degree.values():
          accum_res_iter(resources_used, data_type)
    for module in ('compute',):
      for var in self.instances[module].values():
        accum_res_iter(resources_used, var.values())
    for module in ('forward',):
      for degree in self.instances[module].values():
        for data_type in degree.values():
          for fifo_depth in data_type.values():
            accum_res_iter(resources_used,
                             fifo_depth)
    for resource in resources:
      total = resources_used[resource]
      ground_truth = self.resources_instances[resource]
      if total != ground_truth:
        msg = '%s shall add up to %d, got %d' % (
          resource, ground_truth, total)
        raise ReportSemanticError(msg)
    return True

class XilinxPostRoutingReport():
  def __init__(self, rpt_file):
    for line in rpt_file:
      line_splited = line.split('|')
      if len(line_splited) > 1 and 'Name' in line_splited[1]:
        self.resources_used = collections.OrderedDict(
          (resource.strip(), None)
          for resource in line_splited[2:-1])
      elif (hasattr(self, 'resources_used') and
          len(line_splited) == len(self.resources_used)+3 and
          'Used Resources' in line):
        iterator = iter(line_splited[2:-1])
        for res in self.resources_used:
          self.resources_used[res] = next(iterator).strip().split()[0]

  def __str__(self):
    return json.dumps(self.__dict__, indent=2, sort_keys=True)

class ReportSemanticError(Exception):
  pass

def accum_res(total, delta, coefficient=1):
  for resource in total:
    total[resource] += delta[resource] * coefficient
def accum_res_iter(total, deltas, coefficient=1):
  for delta in deltas:
    accum_res(total, delta, coefficient)

def main():
  for file_name in sys.argv[1:]:
    with open(file_name) as f:
      if file_name.endswith('kernel_util_routed.rpt'):
        rpt = XilinxPostRoutingReport(f)
        print(rpt)
      elif file_name.endswith('_csynth.rpt'):
        rpt = XilinxHLSReport(f)
        rpt.check()
        print(json.dumps({'resources_hls_used': rpt.resources_used},
                 indent=2))

if __name__ == '__main__':
  main()



import collections
import itertools
import logging

from haoda import ir, util
from haoda.ir import arithmetic

from soda import grammar, mutator, visitor

_logger = logging.getLogger().getChild(__name__)


def inline(stencil):
  """Inline statements that are only referenced once.
  """
  if not stencil.local_stmts:
    return stencil

  refs = {}  # type: Dict[str, Set[Tuple[ir.Ref, ir.LocalOrOutputStmt]]]
  for stmt in itertools.chain(stencil.local_stmts, stencil.output_stmts):
    for var_name, ref_list in visitor.get_load_dict(stmt).items():
      if var_name in stencil.input_names or var_name == stmt.name:
        continue
      refs.setdefault(var_name,
                      set()).update(zip(ref_list, itertools.repeat(stmt)))

  refs = {
      name: next(iter(ref_set))
      for name, ref_set in refs.items()
      if len(ref_set) == 1
  }
  if not refs:
    return stencil

  # sort loads to avoid referencing wrong stmt
  local_stmt_table = {
      stmt.name: idx for idx, stmt in enumerate(stencil.local_stmts)
  }
  ref_queue = collections.deque(list(refs.items()))
  sorted_refs = []  # type: List[Tuple[ir.Ref, ir.LocalOrOutputStmt]]
  while ref_queue:
    var_name, (ref, load_stmt) = ref_queue.popleft()
    store_stmt = stencil.local_stmts[local_stmt_table[ref.name]]
    accessed_vars = {ref.name for ref in visitor.get_load_set(store_stmt)}
    queued_vars = {var_name for var_name, _ in ref_queue}
    _logger.debug('stmt to be removed: %s', store_stmt)
    _logger.debug('accessed vars: %s', util.lst2str(accessed_vars))
    _logger.debug('queued vars %s', util.lst2str(queued_vars))
    if accessed_vars & queued_vars:
      ref_queue.append((var_name, (ref, load_stmt)))
    else:
      sorted_refs.append((var_name, (ref, load_stmt)))

  for var_name, (ref, load_stmt) in sorted_refs:
    idx, store_stmt = {
        stmt.name: (idx, stmt) for idx, stmt in enumerate(stencil.local_stmts)
    }[var_name]
    offset = tuple(a - b for a, b in zip(store_stmt.ref.idx, ref.idx))
    ref = mutator.shift(store_stmt.ref, offset)
    lets = tuple(mutator.shift(let, offset) for let in store_stmt.let)
    expr = mutator.shift(store_stmt.expr, offset)
    _logger.info('`%s` is referenced only once, replace with `%s`', ref, expr)
    replace_load = lambda obj, args: args[1] if obj == args[0] else obj
    # TODO: resolve let variable name conflicts
    load_stmt.let = lets + tuple(
        let.visit(replace_load, (ref, expr)) for let in load_stmt.let)
    load_stmt.expr = load_stmt.expr.visit(replace_load, (ref, expr))
    del stencil.local_stmts[idx]

  # invalidate cached_property
  stencil.__dict__.pop('symbol_table', None)
  stencil.__dict__.pop('local_names', None)
  stencil.__dict__.pop('local_types', None)

  for stmt in itertools.chain(stencil.local_stmts, stencil.output_stmts):
    _logger.debug('simplify  : %s', stmt)
    stmt.expr = arithmetic.simplify(stmt.expr)
    stmt.let = arithmetic.simplify(stmt.let)
    _logger.debug('simplified:  %s', stmt)
  return inline(stencil)


def inline2(stencil):
  """Inline statements that are referenced by only one other statement.
  """
  if not stencil.local_stmts:
    return stencil

  refs = collections.OrderedDict(
  )  # type: Dict[str, Dict[ir.LocalOrOutputStmt, List[ir.Ref]]]
  for stmt in itertools.chain(stencil.local_stmts, stencil.output_stmts):
    for var_name, ref_list in visitor.get_load_dict(stmt).items():
      if var_name in stencil.input_names or var_name == stmt.name:
        continue
      refs.setdefault(var_name,
                      collections.OrderedDict()).setdefault(stmt,
                                                            []).extend(ref_list)

  refs = {
      name: next(iter(ref_dict.items()))
      for name, ref_dict in refs.items()
      if len(ref_dict) == 1 and len(
          visitor.get_load_set(
              {stmt.name: stmt.expr
               for stmt in stencil.local_stmts}[name])) == 1
  }
  for name, (stmt, ref_list) in refs.items():
    _logger.info(
        'name: %s stmt: %s ref_list: %s', name, stmt.name,
        util.lst2str(
            visitor.get_load_set(
                {stmt.name: stmt.expr for stmt in stencil.local_stmts}[name])))
  if not refs:
    return stencil

  # sort loads to avoid referencing wrong stmt
  local_stmt_table = {
      stmt.name: idx for idx, stmt in enumerate(stencil.local_stmts)
  }
  ref_queue = collections.deque(list(refs.items()))
  sorted_refs = []  # type: List[Tuple[ir.Ref, ir.LocalOrOutputStmt]]
  while ref_queue:
    var_name, (load_stmt, ref_list) = ref_queue.popleft()
    store_stmt = stencil.local_stmts[local_stmt_table[ref_list[0].name]]
    accessed_vars = {ref.name for ref in visitor.get_load_set(store_stmt)}
    queued_vars = {var_name for var_name, _ in ref_queue}
    _logger.debug('stmt to be removed: %s', store_stmt)
    _logger.debug('accessed vars: %s', util.lst2str(accessed_vars))
    _logger.debug('queued vars %s', util.lst2str(queued_vars))
    if accessed_vars & queued_vars:
      ref_queue.append((var_name, (load_stmt, ref_list)))
    else:
      sorted_refs.append((var_name, (load_stmt, ref_list)))

  for var_name, (load_stmt, ref_list) in sorted_refs:
    idx, store_stmt = {
        stmt.name: (idx, stmt) for idx, stmt in enumerate(stencil.local_stmts)
    }[var_name]
    ref_table = {}
    for ref in ref_list:
      offset = tuple(a - b for a, b in zip(store_stmt.ref.idx, ref.idx))
      ref = mutator.shift(store_stmt.ref, offset)
      lets = tuple(mutator.shift(let, offset) for let in store_stmt.let)
      expr = mutator.shift(store_stmt.expr, offset)
      _logger.info('`%s` is referenced only once by stmt %s, replace with `%s`',
                   ref, load_stmt.name, expr)
      ref_table[ref] = expr
    replace_load = lambda obj, args: args.get(obj, obj)
    # TODO: resolve let variable name conflicts
    load_stmt.let = lets + tuple(
        let.visit(replace_load, ref_table) for let in load_stmt.let)
    load_stmt.expr = load_stmt.expr.visit(replace_load, ref_table)
    del stencil.local_stmts[idx]

  # invalidate cached_property
  stencil.__dict__.pop('symbol_table', None)
  stencil.__dict__.pop('local_names', None)
  stencil.__dict__.pop('local_types', None)

  for stmt in itertools.chain(stencil.local_stmts, stencil.output_stmts):
    _logger.debug('simplify  : %s', stmt)
    stmt.expr = arithmetic.simplify(
        arithmetic.base.reverse_distribute(stmt.expr))
    stmt.let = arithmetic.simplify(
        tuple(map(arithmetic.base.reverse_distribute, stmt.let)))
    _logger.debug('simplified:  %s', stmt)
  return inline2(stencil)


REBALANCE_THRESHOLDS = {
    ir.Type('float'): 32,
}


def rebalance(stencil):
  """Rebalance the generated code to improve codegen speed.

  This function modifies stencil in-place. Long reduction expressions will be
  divided into groups based on the number of operations. The thresholds are
  defined in REBALANCE_THRESHOLDS.

  Arg:
    stencil: The Stencil object to modify.
  Returns:
    The modified Stencil object.
  """
  for stmt in itertools.chain(stencil.local_stmts, stencil.output_stmts):
    if stmt.haoda_type not in REBALANCE_THRESHOLDS:
      continue
    if isinstance(stmt.expr, ir.AddSub) and set(stmt.expr.operator) == {'+'}:
      reduction = []
      for operator, operand in zip(('+',) + getattr(stmt.expr, 'operator'),
                                   getattr(stmt.expr, 'operand')):
        if isinstance(operand, ir.MulDiv) and operand.operator == ('*',):
          opds = getattr(operand, 'operand')
          if isinstance(opds[0], ir.AddSub):
            reduction.append((opds[1], opds[0]))
          elif isinstance(opds[1], ir.AddSub):
            reduction.append((opds[0], opds[1]))
          else:
            reduction.append((None, operand))
        else:
          reduction.append((None, operand))

      get_num_items = lambda x: 1 if x[0] is None else len(x[1].operand)
      reduction.sort(key=get_num_items, reverse=True)

      num_items = 0
      reductions = [[]]
      for coeff, opds in reduction:
        if num_items + get_num_items(
            (coeff, opds)) > REBALANCE_THRESHOLDS[stmt.haoda_type]:
          reductions.append([])
          num_items = 0
        reductions[-1].append((coeff, opds))
        num_items += get_num_items((coeff, opds))
      if len(reductions) == 1:
        continue
      _logger.info('stmt %s has too many operations, breaking\'em into %d ',
                   stmt.name, len(reductions))
      new_stmts = []
      new_exprs = []
      for reduction in reductions:
        new_operators = []
        new_operands = []
        for coeff, opds in reduction:
          new_operators.append('+')
          if coeff is None:
            new_operands.append(opds)
          else:
            new_operands.append(
                ir.MulDiv(operator=('*',), operand=(opds, coeff)))
        new_exprs.append(
            stencil.propagate_type(
                ir.AddSub(operator=tuple(new_operators[1:]),
                          operand=tuple(new_operands))))
      for new_expr in new_exprs[:-1]:
        new_stmt_name = stencil.new_cr_var()
        new_stmts.append(
            grammar.LocalStmt(ref=ir.Ref(name=new_stmt_name,
                                         lat=None,
                                         idx=(0,) * len(stmt.ref.idx)),
                              haoda_type=new_expr.haoda_type,
                              expr=new_expr,
                              let=stmt.let,
                              stencil=stencil))
        _logger.debug('new stmt: %s', new_stmts[-1])
      stencil.local_stmts.extend(new_stmts)
      stmt.expr = ir.AddSub(
          operator=new_exprs[-1].operator + ('+',) * len(new_stmts),
          operand=new_exprs[-1].operand + tuple(stmt.ref for stmt in new_stmts))

      # invalidate cached_property
      stencil.__dict__.pop('symbol_table', None)
      stencil.__dict__.pop('local_names', None)
      stencil.__dict__.pop('local_types', None)

      _logger.debug('stencil after rebalancing: \n  %s',
                    str(stencil).replace('\n', '\n  '))

      return rebalance(stencil)
  return stencil



import functools
import operator

COORDS_TILED = 'xyzw'
COORDS_IN_TILE = 'ijkl'
COORDS_IN_ORIG = 'pqrs'
MAX_DRAM_BANK = 4

def serialize(vec, tile_size):
  return sum((vec[i]*functools.reduce(operator.mul, tile_size[:i])
        for i in range(1, len(tile_size))),
         vec[0])

def serialize_iter(iterative, tile_size):
  return [serialize(x, tile_size) for x in iterative]

def deserialize(offset, tile_size):
  return tuple(deserialize_generator(offset, tile_size))

def deserialize_generator(offset, tile_size):
  for size in tile_size[:-1]:
    yield offset % size
    offset = offset // size
  yield offset



import shutil
import sys
import tempfile

from soda import core
from soda.codegen.intel import ocl_kernel as kernel


def add_arguments(parser):
  parser.add_argument(
      '--iocl-kernel',
      type=str,
      dest='iocl_kernel',
      metavar='file',
      help='Intel OpenCL kernel code for the Intel FPGA OpenCL flow')


def print_code(stencil: core.Stencil, args):
  if args.iocl_kernel is not None:
    with tempfile.TemporaryFile(mode='w+') as tmp:
      kernel.print_code(stencil, tmp)
      tmp.seek(0)
      if args.iocl_kernel == '-':
        shutil.copyfileobj(tmp, sys.stdout)
      else:
        with open(args.iocl_kernel, 'w') as iocl_kernel:
          shutil.copyfileobj(tmp, iocl_kernel)



import collections
import copy
import logging

import cached_property

from haoda import ir
from haoda import util
from soda import grammar
import soda.util

_logger = logging.getLogger().getChild(__name__)

class Tensor():
  """A tensor that corresponse to an input, local, or output.

  This class is used in the high-level DAG for stencil dependency analysis.
  Each tensor either is an input tensor, or has at least 1 parent tensor, which
  will be used to generate this tensor. Meanwhile, each tensor either is an
  output tensor, or has at least 1 child tensor, which will be computed using
  this tensor.

  Attributes:
    haoda_type: str, type of the tensor element.
    parents: Dict from str of name of Tensor to Tensor.
    children: Dict from str of name of Tensor to Tensor.
    st_ref: Ref, name, index, and latency stored.
    offset: int, shift offset in terms of data elements
    lets: Lets of computation.
    expr: Expr of computation.
    ld_refs: Dict from str of name to dict of Ref loaded.

  Property:
    name: str, unique in each SODA program.
    st_offset: int, stencil offset in terms of data elements.
    st_idx, Tuple of int, the index referenced by its parent stage.
    ld_indices: Dict from str of name to dict of accessed indices of the input.
    ld_offsets: Dict from str of name to dict of offsets of the input.
  """
  def __init__(self, stmt, tile_size):
    self.haoda_type = stmt.haoda_type
    self._tile_size = tile_size
    if isinstance(stmt, grammar.LocalStmtOrOutputStmt):
      self.st_ref = copy.copy(stmt.ref)
      self.st_ref.parent = self
      self.lets = stmt.let
      self.expr = stmt.expr
    elif isinstance(stmt, grammar.InputStmt):
      self._name = stmt.name
      self.st_ref = None
      self.lets = []
      self.expr = None
    else:
      raise util.InternalError('cannot initialize a Tensor from %s' %
                               type(stmt))
    _logger.debug('tensor initialized from stmt `%s`', stmt)
    # pylint: disable=protected-access
    _logger.debug('                   at tx position %d', stmt._tx_position)

    # these fields are to be set externally
    self.parents = collections.OrderedDict()
    self.children = collections.OrderedDict()
    self.ld_refs = collections.OrderedDict()

  @property
  def name(self):
    if self.st_ref is not None:
      return self.st_ref.name
    return self._name

  @property
  def st_idx(self):
    if self.st_ref is not None:
      return self.st_ref.idx
    return (0,)*len(self._tile_size)

  @property
  def st_offset(self):
    return soda.util.serialize(self.st_idx, self._tile_size)

  @cached_property.cached_property
  def ld_indices(self):
    return collections.OrderedDict(
        (name, collections.OrderedDict((ref.idx, ref) for ref in refs))
        for name, refs in self.ld_refs.items())

  @cached_property.cached_property
  def ld_offsets(self):
    return collections.OrderedDict(
      (name, collections.OrderedDict(
        (soda.util.serialize(ref.idx, self._tile_size), ref) for ref in refs))
      for name, refs in self.ld_refs.items())

  @property
  def c_type(self):
    return self.haoda_type.c_type

  def propagate_type(self):
    if self.expr is None:
      return

    var_types = {}
    # pylint: disable=access-member-before-definition
    for let in self.lets:
      var_types[let.name] = let.haoda_type

    def visit_haoda_type(obj, args):
      if obj.haoda_type is None:
        if isinstance(obj, ir.Var):
          obj.haoda_type = var_types[obj.name]
      return obj

    self.lets = tuple(_.visit(visit_haoda_type) for _ in self.lets)
    self.expr = self.expr.visit(visit_haoda_type)
    self.st_ref = self.st_ref.visit(visit_haoda_type)

  def mutate(self, callback, args=None):
    self.lets = tuple(_.visit(callback, args) for _ in self.lets)
    self.expr = self.expr.visit(callback, args)
    self.st_ref = self.st_ref.visit(callback, args)

  def visit_loads(self, callback, args=None):
    for let in self.lets:
      let.visit(callback, args)
    self.expr.visit(callback, args)

  def __str__(self):
    return '''Tensor
  {haoda_type}: {name} = {expr}
  store: {st_ref}
  parents: {parents}
  children: {children}'''.format(
      name=self.name, haoda_type=self.haoda_type, expr=self.expr,
      parents=util.idx2str(self.parents), children=util.idx2str(self.children),
      st_ref=str(self.st_ref))

  def is_output(self):
    return len(self.children) == 0

  def is_input(self):
    return len(self.parents) == 0

  def is_producer(self):
    return not self.is_output()

  def is_consumer(self):
    return not self.is_input()



import collections
import logging
from typing import Dict, List, TextIO

from haoda import ir, util
from soda import core

_logger = logging.getLogger().getChild(__name__)


def print_code(stencil: core.Stencil, output_file: TextIO) -> None:
  """Prints the top-level code with the given arguments.

  Prints the OpenCL kernels with proper pragmas and channel declarations and
  references.

  Args:
    stencil: Stencil object to print.
    output_file: TextIO to write to.
  """
  _logger.info('generate kernel code as %s' % output_file.name)
  printer = util.CppPrinter(output_file)
  println = printer.println

  println('#include <ihc_apint.h>', indent=0)
  println()

  println('#pragma OPENCL EXTENSION cl_intel_channels : enable', indent=0)
  println()

  # internal fifos
  super_source = stencil.dataflow_super_source
  for node in super_source.tpo_valid_node_gen():
    for fifo in node.fifos:
      println(f'channel {fifo.cl_type} {fifo.cl_expr};')

  println()

  # in generated bitstream, kernels are sorted in alphabetical order
  # SODA relies on the correct ordering for memory channels of each tensor
  # so here we make sure the kernel names in alphabetical order
  width = len(str(sum(1 for _ in super_source.tpo_valid_node_gen()) - 1))
  instance_idx: Dict[int, int] = collections.defaultdict(int)
  overall_idx = 0
  for node in super_source.tpo_valid_node_gen():
    module_trait, module_trait_id = super_source.module_table[node]
    print_kernel(
        f'{stencil.app_name}_{overall_idx:0{width}}_'
        f'module_{module_trait_id}_'
        f'instance_{instance_idx[module_trait_id]}',
        printer,
        node,
        module_trait,
        module_trait_id,
        burst_width=stencil.burst_width,
    )
    instance_idx[module_trait_id] += 1
    overall_idx += 1
    println()


def print_kernel(name: str,
                 printer: util.Printer,
                 node: ir.Module,
                 module_trait: ir.ModuleTrait,
                 module_trait_id: int,
                 burst_width: int = 256) -> None:
  if node.dram_reads and node.dram_writes:
    raise ValueError('cannot read and write DRAM in the same module')
  println = printer.println

  # print I/O info
  for port, arg in zip(module_trait.loads, node.input_fifos):
    println('//  input <{0.cl_type}> <- {1}'.format(port, arg))
  for expr, arg in zip(module_trait.exprs, node.output_fifos):
    println('// output <{}> -> {}'.format(expr.cl_type, arg))

  # print kernel function
  printer.println('__kernel')
  kernel_attrs = ['reqd_work_group_size(1, 1, 1)', 'max_global_work_dim(0)']

  def print_kernel_attrs():
    for attr in kernel_attrs:
      println('__attribute(({}))'.format(attr))

  if node.dram_reads or node.dram_writes:
    params = [
        '__global {} {}* restrict {}'.format(
            '__attribute((buffer_location("HBM{}")))'.format(bank),
            dram_ref.haoda_type.get_cl_vec_type(burst_width),
            util.get_port_name(dram_ref.var, bank))
        for dram_ref, bank in node.dram_reads or node.dram_writes
    ]
    params.append('ulong coalesced_data_num')
    kernel_attrs.append('uses_global_work_offset(0)')
    print_kernel_attrs()
    printer.print_func(name='void {}'.format(name), params=params, align=0)
  else:
    kernel_attrs.append('autorun')
    print_kernel_attrs()
    println('void {}()'.format(name))
  printer.do_scope(name)

  # prepare for any DelayedRef
  def get_delays(obj: ir.Node, delays: List[ir.DelayedRef]) -> ir.Node:
    if isinstance(obj, ir.DelayedRef):
      delays.append(obj)
    return obj

  delays = []  # type: List[ir.DelayedRef]
  for let in module_trait.lets:
    let.visit(get_delays, delays)
  for expr in module_trait.exprs:
    expr.visit(get_delays, delays)
  _logger.debug('delays: %s', delays)

  # inter-iteration declarations
  for delay in delays:
    println(delay.cl_buf_decl)
    println(delay.cl_ptr_decl)

  def mutate_dram_ref(obj: ir.Node, kwargs: Dict[str, int]) -> ir.Node:
    if isinstance(obj, ir.DRAMRef):
      dram_throughput = len(obj.dram) * burst_width // obj.width_in_bits
      if node.dram_reads:
        output_count = len({x[0].var for x in node.dram_reads})
        fifo_throughput = len(node.output_fifos) // output_count
      else:
        input_count = len({x[0].var for x in node.dram_writes})
        fifo_throughput = len(node.input_fifos) // input_count
      if dram_throughput != fifo_throughput:
        raise NotImplementedError(f'memory throughput {dram_throughput} != '
                                  f'processing throughput {fifo_throughput}')

      coalescing_idx = kwargs['coalescing_idx']
      unroll_factor = kwargs['unroll_factor']
      elem_idx = coalescing_idx * unroll_factor + obj.offset
      return ir.Var(
          name='{buf}[{idx}]'.format(
              buf=obj.dram_buf_name(obj.dram[elem_idx % len(obj.dram)]),
              idx=elem_idx // len(obj.dram),
          ),
          idx=(),
      )
    if isinstance(obj, ir.Let) and isinstance(obj.name, ir.DRAMRef):
      return ir.Var(name='{} = {};'.format(
          obj.name.visit(mutate_dram_ref, kwargs), obj.expr.cl_expr),
                    idx=())
    return obj

  if node.dram_reads or node.dram_writes:
    loop_args = 'ulong i = 0', 'i < coalesced_data_num', '++i'
  else:
    loop_args = '', '', ''

  if delays:
    println('#pragma ivdep', indent=0)
  with printer.for_(*loop_args):
    # print DelayedRef (if any)
    for delay in delays:
      println('const {} {};'.format(delay.cl_type, delay.c_buf_load))

    # print load from DRAM (if any)
    for dram_ref, bank in node.dram_reads:
      println('{gentype} {buf}[{n}];'.format(
          gentype=dram_ref.cl_type,
          n=burst_width // dram_ref.width_in_bits,
          buf=dram_ref.dram_buf_name(bank),
      ))
    for dram_ref, bank in node.dram_reads:
      println('vstore{n}({vec_ptr}[i], 0, {buf});'.format(
          vec_ptr=util.get_port_name(dram_ref.var, bank),
          n=burst_width // dram_ref.width_in_bits,
          buf=dram_ref.dram_buf_name(bank),
      ))

    # read from FIFOs
    for port, arg in zip(module_trait.loads, node.input_fifos):
      println('{0.cl_type} {0.ref_name} = read_channel_intel({1});'.format(
          port, arg))

    # declare buffer for DRAM writes
    for dram_ref, bank in node.dram_writes:
      n = burst_width // dram_ref.width_in_bits
      buf = dram_ref.dram_buf_name(bank)
      println('{gentype} {buf}[{n}];'.format(
          gentype=dram_ref.cl_type,
          n=n,
          buf=buf,
      ))

    mutate_kwargs = {
        'coalescing_idx': 0,
        'unroll_factor': len(node.input_fifos) or len(node.output_fifos),
    }

    # print Let (if any)
    for let in module_trait.lets:
      println(let.visit(mutate_dram_ref, mutate_kwargs).cl_expr)

    for expr, arg in zip(module_trait.exprs, node.output_fifos):
      println('write_channel_intel({}, {});'.format(
          arg,
          expr.visit(mutate_dram_ref, mutate_kwargs).cl_expr))

    # update DelayedRef (if any)
    for delay in delays:
      println(delay.c_buf_store)
      println('{} = {};'.format(delay.ptr, delay.cl_next_ptr_expr))

    # print store to DRAM (if any)
    for dram_ref, bank in node.dram_writes:
      println('{vec_ptr}[i] = vload{n}(0, {buf});'.format(
          vec_ptr=util.get_port_name(dram_ref.var, bank),
          n=burst_width // dram_ref.width_in_bits,
          buf=dram_ref.dram_buf_name(bank),
      ))

  printer.un_scope()  # end of kernel function
  _logger.debug('printing: %s', module_trait)



import argparse
import shutil
import sys
import tempfile

from soda import core
from soda.codegen.frt import host


def add_arguments(parser: argparse.ArgumentParser) -> None:
  parser.add_argument('--frt-host',
                      type=str,
                      dest='frt_host',
                      metavar='file',
                      help='FPGA runtime host code')


def print_code(stencil: core.Stencil, args: argparse.Namespace) -> None:
  if args.frt_host is not None:
    with tempfile.TemporaryFile(mode='w+') as tmp:
      host.print_code(stencil, tmp)
      tmp.seek(0)
      if args.frt_host == '-':
        shutil.copyfileobj(tmp, sys.stdout)
      else:
        with open(args.frt_host, 'w') as frt_host:
          shutil.copyfileobj(tmp, frt_host)



import functools
import logging
import operator
from typing import List, TextIO

import soda.tensor
from haoda import ir, util
from soda import core, grammar
from soda import util as soda_util

logger = logging.getLogger().getChild(__name__)

STENCIL_DIM_FMT = util.MetaFmt('kStencilDim%s')
TYPE_FMT = util.MetaFmt('Type_%s')
WIDTH_FMT = util.MetaFmt('kWidth_%s')


def print_header(printer):
  # C headers
  for header in ('assert', 'float', 'math', 'stdbool', 'stddef', 'stdint',
                 'stdio', 'stdlib', 'string'):
    printer.println(f'#include <c{header}>')
  printer.println()

  # C++ headers
  for header in ('algorithm', 'iomanip', 'iostream', 'memory', 'random',
                 'regex', 'string', 'unordered_map', 'vector'):
    printer.println(f'#include <{header}>')
  printer.println()

  # Other system headers
  for header in ('frt',):
    printer.println(f'#include <{header}.h>')
  printer.println()

  # using declarations
  for name in ('clog', 'endl', 'regex', 'regex_match', 'string', 'unique_ptr',
               'unordered_map', 'vector'):
    printer.println(f'using std::{name};')
  printer.println()


def print_func(printer: util.CppPrinter, stencil: soda.core.Stencil):
  stmts = stencil.input_stmts + stencil.output_stmts

  # factories for meta variables
  data_fmt = util.MetaFmt('var_%s_ptr')
  extent_fmt = util.MetaFmt('var_%s_extent')
  stride_fmt = util.MetaFmt('var_%s_stride')
  min_fmt = util.MetaFmt('var_%s_min')

  # print function signature
  params: List[str] = []
  for stmt in stmts + stencil.param_stmts:
    prefix = 'const ' if isinstance(stmt, grammar.InputStmt) else ''
    params.extend((f'{prefix}{TYPE_FMT[stmt.name]}* {data_fmt[stmt.name]}',
                   f'const int32_t {extent_fmt[stmt.name]}[{stencil.dim}]',
                   f'const int32_t {stride_fmt[stmt.name]}[{stencil.dim}]',
                   f'const int32_t {min_fmt[stmt.name]}[{stencil.dim}]'))

  tile_size_fmt = util.MetaFmt('tile_size_%s')
  params.extend((
      'const char* bitstream',
      f'const int burst_width = {stencil.burst_width}',
      *(f'const int {tile_size_fmt[d]} = {stencil.tile_size[d]}'
        for d in range(stencil.dim - 1)),
      f'const int unroll_factor = {stencil.unroll_factor}',
  ))
  printer.print_func(name=f'int {stencil.app_name}', params=params, align=0)
  printer.do_scope()

  printer.printlns(
      '// load bitstream',
      'auto instance = fpga::Instance(bitstream);',
      'auto args_info = instance.GetArgsInfo();'
      '',
  )

  bank_count_fmt = util.MetaFmt('bank_count_%s')
  regex_fmt = util.MetaFmt('regex_%s')
  elem_count_per_cycle_fmt = util.MetaFmt('elem_count_per_cycle_%s')
  tile_count_fmt = util.MetaFmt('tile_count_dim_%d')
  printer.printlns(
      '// find out how many banks are used for each tensor',
      *(f'int {bank_count_fmt[x.name]} = 0;' for x in stmts),
      *(f'const regex {regex_fmt[x.name]}'
        f'(R"(^bank_\\d+_{x.name}$)");' for x in stmts),
  )
  with printer.for_('const auto& arg', 'args_info'):
    printer.printlns(f'if (regex_match(arg.name, {regex_fmt[x.name]})) '
                     f'++{bank_count_fmt[x.name]};' for x in stmts)
  printer.printlns(
      '',
      ('auto round_up = [](int64_t a, int64_t b) -> int64_t '
       '{ return ((a - 1) / b + 1) * b; };'),
      '',
      '// some run-time constants',
      *(f'const int {elem_count_per_cycle_fmt[x.name]} = '
        f'burst_width / {WIDTH_FMT[x.name]} * {bank_count_fmt[x.name]};'
        for x in stmts),
  )
  for d in range(stencil.dim - 1):
    printer.println(f'int32_t {tile_count_fmt[d]} = '
                    f'({extent_fmt[stencil.input_names[0]]}[{d}] - '
                    f'{STENCIL_DIM_FMT[d]} + 1 - 1) / ({tile_size_fmt[d]} - '
                    f'{STENCIL_DIM_FMT[d]} + 1) + 1;')
  printer.printlns(
      ('int64_t tile_count = %s;' %
       ' * '.join(f'{tile_count_fmt[d]}' for d in range(stencil.dim - 1))),
      '',
  )

  printer.printlns(
      '// align each linearized tile to multiples of burst_width',
      ('int64_t elem_count_per_tile = %s * '
       f'{extent_fmt[stencil.input_names[0]]}[{stencil.dim - 1}];' %
       ' * '.join(f'{tile_size_fmt[d]}' for d in range(stencil.dim - 1))),
      ('int64_t cycle_count_per_tile = (elem_count_per_tile - 1) / '
       f'{elem_count_per_cycle_fmt[stencil.input_names[0]]} + 1;'),
      ('int64_t elem_count_aligned_per_tile_i = cycle_count_per_tile * '
       f'{elem_count_per_cycle_fmt[stencil.input_stmts[0].name]};'),
      ('int64_t elem_count_aligned_per_tile_o = cycle_count_per_tile * '
       f'{elem_count_per_cycle_fmt[stencil.output_stmts[0].name]};'),
      '',
  )

  printer.println('// calculate size of each buffer')
  buf_size_fmt = util.MetaFmt('buf_size_%s')
  for stmt in stencil.input_stmts:
    printer.println(
        f'int64_t {buf_size_fmt[stmt.name]} = '
        f'(tile_count * elem_count_aligned_per_tile_i + '
        f'round_up(kStencilDistance, {elem_count_per_cycle_fmt[stmt.name]}))'
        f' / {bank_count_fmt[stmt.name]} * sizeof({TYPE_FMT[stmt.name]});')
  for stmt in stencil.output_stmts:
    printer.println(
        f'int64_t {buf_size_fmt[stmt.name]} = '
        f'(tile_count * elem_count_aligned_per_tile_o + '
        f'round_up(kStencilDistance, {elem_count_per_cycle_fmt[stmt.name]}))'
        f' / {bank_count_fmt[stmt.name]} * sizeof({TYPE_FMT[stmt.name]});')
  printer.println()

  printer.println('// allocate memory for each buffer')
  buf_fmt = util.MetaFmt('buf_%s')
  for stmt in stmts:
    printer.printlns(
        (f'vector<unique_ptr<{TYPE_FMT[stmt.name]}, decltype(&free)>> '
         f'{buf_fmt[stmt.name]};'),
        f'{buf_fmt[stmt.name]}.reserve({bank_count_fmt[stmt.name]});',
    )
    with printer.for_('int bank = 0', f'bank < {bank_count_fmt[stmt.name]}',
                      '++bank'):
      printer.println(
          f'{buf_fmt[stmt.name]}.emplace_back('
          f'static_cast<{TYPE_FMT[stmt.name]}*>(aligned_alloc('
          f'4096, round_up({buf_size_fmt[stmt.name]}, 4096))), &free);')
  printer.println()

  printer.println('// tiling')
  for dim in range(stencil.dim - 2, -1, -1):
    printer.println(f'for(int32_t tile_index_dim_{dim} = 0; '
                    f'tile_index_dim_{dim} < {tile_count_fmt[dim]}; '
                    f'++tile_index_dim_{dim})')
    printer.do_scope()
    printer.println(f'int32_t actual_tile_size_dim_{dim} = '
                    f'(tile_index_dim_{dim}=={tile_count_fmt[dim]}-1) ? '
                    f'{extent_fmt[stencil.input_names[0]]}[{dim}] - '
                    f'({tile_size_fmt[dim]} - {STENCIL_DIM_FMT[dim]} + 1) * '
                    f'tile_index_dim_{dim} : {tile_size_fmt[dim]};')

  printer.println('#pragma omp parallel for', 0)
  var = soda_util.COORDS_IN_TILE[stencil.dim - 1]
  printer.println(
      f'for(int32_t {var} = 0; '
      f'{var} < {extent_fmt[stencil.input_names[0]]}[{stencil.dim - 1}]; '
      f'++{var})')
  printer.do_scope()
  for dim in range(stencil.dim - 2, -1, -1):
    printer.println('for(int32_t {0} = 0; {0} < actual_tile_size_dim_{1}; '
                    '++{0})'.format(soda_util.COORDS_IN_TILE[dim], dim))
    printer.do_scope()

  printer.printlns(
      ('// (%s) is coordinates in tiled image' %
       ', '.join(soda_util.COORDS_TILED)),
      ('// (%s) is coordinates in original image' %
       ', '.join(soda_util.COORDS_IN_ORIG)),
      '// (%s) is coordinates in a tile' % ', '.join(soda_util.COORDS_IN_TILE),
  )
  offset_in_tile = ' + '.join(
      '%s%s' % (soda_util.COORDS_IN_TILE[x], ''.join(f' * {tile_size_fmt[d]}'
                                                     for d in range(x)))
      for x in range(stencil.dim))
  stmt = stencil.input_stmts[0]
  printer.printlns(
      (f'int32_t burst_index = ({offset_in_tile}) / '
       f'{elem_count_per_cycle_fmt[stmt.name]};'),
      (f'int32_t burst_residue = ({offset_in_tile}) % '
       f'{elem_count_per_cycle_fmt[stmt.name]};'),
  )
  for dim in range(stencil.dim - 1):
    printer.println(
        f'int32_t {soda_util.COORDS_IN_ORIG[dim]} = tile_index_dim_{dim} * '
        f'({tile_size_fmt[dim]} - {STENCIL_DIM_FMT[dim]}) + '
        f'{soda_util.COORDS_IN_TILE[dim]};')
  printer.printlns(
      ('int32_t %s = %s;' % (soda_util.COORDS_IN_ORIG[stencil.dim - 1],
                             soda_util.COORDS_IN_TILE[stencil.dim - 1])),
      (f'int64_t tiled_offset = (%s) * elem_count_aligned_per_tile_i + '
       f'burst_index * {elem_count_per_cycle_fmt[stmt.name]} + burst_residue;' %
       ' + '.join('%stile_index_dim_%d' % (''.join(f'{tile_count_fmt[d]} * '
                                                   for d in range(x)), x)
                  for x in range(stencil.dim - 1))),
      ('int64_t original_offset = %s;' %
       ' + '.join(f'%s * {stride_fmt[stencil.input_names[0]]}[%d]' %
                  (soda_util.COORDS_IN_ORIG[x], x)
                  for x in range(stencil.dim))),
  )
  printer.printlns(f'{buf_fmt[x]}'
                   f'[tiled_offset % {bank_count_fmt[x]}].get()'
                   f'[tiled_offset / {bank_count_fmt[x]}] = '
                   f'{data_fmt[x]}[std::max(int64_t(0), original_offset - '
                   f'{stencil.tensors[x].produce_offset})];'
                   for x in stencil.input_names)
  for dim in range(stencil.dim * 2 - 1):
    printer.un_scope()
  printer.println()

  for d in range(stencil.dim - 1):
    printer.println(
        f'clog << "INFO: tile_count[{d}] = " << {tile_count_fmt[d]} '
        f'<< ", tile_size[{d}] = " << {tile_size_fmt[d]} << endl;')
  for name in stencil.input_names + stencil.output_names:
    for item in 'extent', 'stride', 'min':
      fmt = locals()[item + '_fmt']
      printer.println(
          'clog << "INFO: %s" << endl;' %
          ', '.join(f'{name}.{item}[{d}] = " << {fmt[name]}[{d}] << "'
                    for d in range(stencil.dim)))
  printer.println()

  stmt = stencil.input_stmts[0]
  printer.printlns(
      ('int64_t tile_data_count = '
       f'((int64_t({extent_fmt[stmt.name]}[{stencil.dim - 1}])%s - 1) / '
       f'{elem_count_per_cycle_fmt[stmt.name]} + 1) * '
       f'{elem_count_per_cycle_fmt[stmt.name]} / '
       'unroll_factor;' %
       (''.join(f' * {tile_size_fmt[d]}' for d in range(stencil.dim - 1)))),
      ('int64_t cycle_count = '
       f'((int64_t({extent_fmt[stmt.name]}[{stencil.dim - 1}])%s * %s + '
       f'kStencilDistance - 1) / {elem_count_per_cycle_fmt[stmt.name]} + 1);' %
       (''.join(f' * {tile_size_fmt[d]}' for d in range(stencil.dim - 1)),
        ' * '.join(tile_count_fmt[d] for d in range(stencil.dim - 1)))),
      ('clog << "INFO: tile_data_count = " << tile_data_count '
       '<< ", cycle_count = " << cycle_count << endl;'),
      '',
  )

  printer.println('int arg_idx = 0;')
  iter_fmt = util.MetaFmt('iter_%s')
  for stmt in stmts:
    printer.println(
        f'auto {iter_fmt[stmt.name]} = {buf_fmt[stmt.name]}.begin();')
  with printer.for_('const auto& arg', 'args_info'):
    with printer.if_('arg.name == "coalesced_data_num"'):
      printer.printlns(
          'instance.SetArg(arg_idx, cycle_count);',
          '++arg_idx;',
      )
      for stmt in stmts:
        direction = 'Write' if isinstance(stmt, grammar.InputStmt) else 'Read'
        with printer.elif_(f'regex_match(arg.name, {regex_fmt[stmt.name]})'):
          printer.printlns(
              (f'auto buf = fpga::{direction}Only('
               f'{iter_fmt[stmt.name]}->get(), '
               f'{buf_size_fmt[stmt.name]} / sizeof({TYPE_FMT[stmt.name]}));'),
              'instance.AllocBuf(arg_idx, buf);',
              'instance.SetArg(arg_idx, buf);',
              f'++{iter_fmt[stmt.name]};',
              '++arg_idx;',
          )

  printer.printlns(
      '',
      'instance.WriteToDevice();',
      'instance.Exec();',
      'instance.ReadFromDevice();',
      'instance.Finish();',
      '',
      ('clog << "Load throughput: " << std::setprecision(3) '
       '<< instance.LoadThroughputGbps() << " GB/s" << endl;'),
      ('clog << "Compute latency: " << std::setprecision(3) '
       '<< instance.ComputeTimeSeconds() << " s" << endl;'),
      ('clog << "Store throughput: " << std::setprecision(3) '
       '<< instance.StoreThroughputGbps() <<" GB/s" << endl;'),
      '',
  )

  for dim in range(stencil.dim - 2, -1, -1):
    printer.println(
        f'for(int32_t tile_index_dim_{dim} = 0; tile_index_dim_{dim} < '
        f'{tile_count_fmt[dim]}; ++tile_index_dim_{dim})')
    printer.do_scope()
    printer.println(f'int32_t actual_tile_size_dim_{dim} = '
                    f'(tile_index_dim_{dim} == {tile_count_fmt[dim]}-1) ? '
                    f'{extent_fmt[stencil.input_names[0]]}[{dim}] - '
                    f'({tile_size_fmt[dim]} - {STENCIL_DIM_FMT[dim]} + 1)'
                    f' * tile_index_dim_{dim} : {tile_size_fmt[dim]};')

  overall_stencil_window = core.get_overall_stencil_window(
      stencil.tensors[stencil.input_names[0]],
      stencil.tensors[stencil.output_names[0]])
  overall_stencil_offset = core.get_stencil_window_offset(
      overall_stencil_window)
  overall_stencil_dim = core.get_stencil_dim(overall_stencil_window)
  printer.println('#pragma omp parallel for', 0)
  printer.println('for(int32_t {var} = {}; {var} < '
                  f'{extent_fmt[stencil.output_names[0]]}[{stencil.dim - 1}]'
                  ' - {}; ++{var})'.format(
                      max(0, overall_stencil_offset[stencil.dim - 1]),
                      max(0, (overall_stencil_dim[stencil.dim - 1] - 1 -
                              overall_stencil_offset[stencil.dim - 1])),
                      var=soda_util.COORDS_IN_TILE[stencil.dim - 1]))
  printer.do_scope()
  for dim in range(stencil.dim - 2, -1, -1):
    printer.println(
        'for(int32_t {var} = {}; {var} < actual_tile_size_dim_{} - {}; '
        '++{var})'.format(
            max(0, overall_stencil_offset[dim]),
            dim,
            max(0, overall_stencil_dim[dim] - 1 - overall_stencil_offset[dim]),
            var=soda_util.COORDS_IN_TILE[dim]))
    printer.do_scope()

  printer.printlns(
      ('// (%s) is coordinates in tiled image' %
       ', '.join(soda_util.COORDS_TILED)),
      ('// (%s) is coordinates in original image' %
       ', '.join(soda_util.COORDS_IN_ORIG)),
      '// (%s) is coordinates in a tile' % ', '.join(soda_util.COORDS_IN_TILE),
  )
  offset_in_tile = ' + '.join(
      '%s%s' % (soda_util.COORDS_IN_TILE[x], ''.join(f' * {tile_size_fmt[d]}'
                                                     for d in range(x)))
      for x in range(stencil.dim))
  for dim in range(stencil.dim - 1):
    printer.println(
        f'int32_t {soda_util.COORDS_IN_ORIG[dim]} = tile_index_dim_{dim} '
        f'* ({tile_size_fmt[dim]}-{STENCIL_DIM_FMT[dim]} + 1) + '
        f'{soda_util.COORDS_IN_TILE[dim]};')
  printer.printlns(
      ('int32_t %s = %s;' % (soda_util.COORDS_IN_ORIG[stencil.dim - 1],
                             soda_util.COORDS_IN_TILE[stencil.dim - 1])),
      ('int64_t original_offset = %s;' %
       ' + '.join(f'%s * {stride_fmt[stencil.output_names[0]]}[%d]' %
                  (soda_util.COORDS_IN_ORIG[x], x)
                  for x in range(stencil.dim))),
  )
  for stmt in stencil.output_stmts:
    overall_stencil_window = core.get_overall_stencil_window(
        map(stencil.tensors.get, stencil.input_names),
        stencil.tensors[stmt.name])
    overall_stencil_distance = core.get_stencil_distance(
        overall_stencil_window, stencil.tile_size)
    stencil_offset = overall_stencil_distance - soda_util.serialize(
        core.get_stencil_window_offset(overall_stencil_window),
        stencil.tile_size)
    printer.printlns(
        (f'int32_t burst_index_{stmt.name} = '
         f'({offset_in_tile} + {stencil_offset}) / '
         f'{elem_count_per_cycle_fmt[stmt.name]};'),
        (f'int32_t burst_residue_{stmt.name} = '
         f'({offset_in_tile} + {stencil_offset}) % '
         f'{elem_count_per_cycle_fmt[stmt.name]};'),
        (f'int64_t tiled_offset_{stmt.name} = '
         f'(%s) * elem_count_aligned_per_tile_o + burst_index_{stmt.name} * '
         f'{elem_count_per_cycle_fmt[stmt.name]} + burst_residue_{stmt.name};' %
         ('+'.join('%stile_index_dim_%d' % (''.join(f'{tile_count_fmt[d]} * '
                                                    for d in range(x)), x)
                   for x in range(stencil.dim - 1)))),
        (f'{data_fmt[stmt.name]}[original_offset] = {buf_fmt[stmt.name]}'
         f'[tiled_offset_{stmt.name} % {bank_count_fmt[stmt.name]}].get()'
         f'[tiled_offset_{stmt.name} / {bank_count_fmt[stmt.name]}];'),
    )
  for dim in range(stencil.dim * 2 - 1):
    printer.un_scope()

  printer.println('return 0;')
  printer.un_scope()
  printer.println()


def print_test(printer, stencil):
  stencil.dim = stencil.dim

  printer.printlns(
      '#ifdef SODA_TEST_MAIN',
      '',
      '#include <ap_int.h>',
      '',
      'int main(int argc, char *argv[])',
  )
  printer.do_scope()

  with printer.if_(f'argc < 2 || argc > {stencil.dim + 2}'):
    printer.printlns(
        (r'clog << "Usage: \n  " << argv[0] << " <xclbin> %s" << endl;' %
         ' '.join(map('[input dimension {}]'.format, range(stencil.dim)))),
        'return 2;',
    )

  printer.println('const char* xclbin = argv[1];')
  default_sizes = []
  for d in range(stencil.dim):
    if d != stencil.dim - 1:
      default_sizes.append(str(stencil.tile_size[d]))
    else:
      default_sizes.append(f'soda::app::{STENCIL_DIM_FMT[d]} + 1')
  printer.println(f'int32_t dims[{stencil.dim}] = {{%s}};' %
                  ', '.join(default_sizes))

  for d in range(stencil.dim):
    with printer.if_(f'argc > {d + 2}'):
      printer.println(f'dims[{d}] = std::max(dims[{d}], atoi(argv[{d + 2}]));')
  printer.println()

  data_fmt = util.MetaFmt('data_%s')
  extent_fmt = util.MetaFmt('extent_%s')
  stride_fmt = util.MetaFmt('stride_%s')
  min_fmt = util.MetaFmt('min_%s')
  for tensor in stencil.tensors.values():
    # data
    printer.println(f'vector<{tensor.c_type}> {data_fmt[tensor.name]}(%s);' %
                    (' * '.join('dims[%d]' % x for x in range(stencil.dim))))

    # extent
    printer.println(
        f'int32_t {extent_fmt[tensor.name]}[{stencil.dim}] = {{%s}};' %
        ', '.join(map('dims[{}]'.format, range(stencil.dim))))

    # stride
    stride = []
    for d in range(stencil.dim):
      if d == 0:
        stride.append('1')
      else:
        stride.append(' * '.join(map('dims[{}]'.format, range(d))))
    printer.println(
        f'int32_t {stride_fmt[tensor.name]}[{stencil.dim}] = {{%s}};' %
        ', '.join(stride))

    # min
    printer.println(f'int32_t {min_fmt[tensor.name]}[{stencil.dim}] = {{%s}};' %
                    ', '.join('0' * stencil.dim))

  for param in stencil.param_stmts:
    printer.println(
        f'{param.c_type} {data_fmt[param.name]}%s;' %
        functools.reduce(operator.add, ['[%s]' % x for x in param.size]))
  printer.println()

  if any(x.is_float for x in stencil.input_types):
    printer.printlns(
        'std::default_random_engine generator;',
        'std::uniform_real_distribution<double> distribution(0.0, 1.0);',
        '',
    )

  for name in stencil.input_names:
    printer.println('// initialization can be parallelized with -fopenmp')
    printer.println('#pragma omp parallel for', 0)
    for d in range(0, stencil.dim):
      dim = stencil.dim - d - 1
      printer.println(
          'for(int32_t {var} = 0; {var} < dims[{}]; ++{var})'.format(
              dim, var=soda_util.COORDS_IN_ORIG[dim]))
      printer.do_scope()
    init_val = '+'.join(soda_util.COORDS_IN_ORIG[0:stencil.dim])
    if stencil.symbol_table[name].is_float:
      init_val = 'distribution(generator)'
    printer.println(f'{data_fmt[name]}[%s] = %s;' %
                    (' + '.join(f'%s * {stride_fmt[name]}[%d]' %
                                (soda_util.COORDS_IN_ORIG[d], d)
                                for d in range(stencil.dim)), init_val))
    for d in range(0, stencil.dim):
      printer.un_scope()
    printer.println()

  for param in stencil.param_stmts:
    printer.println('#pragma omp parallel for', 0)
    for dim, size in enumerate(param.size):
      printer.println('for(int32_t {var} = 0; {var}<{}; ++{var})'.format(
          size, var=soda_util.COORDS_IN_ORIG[dim]))
      printer.do_scope()
    printer.println(f'{data_fmt[param.name]}%s = %s;' %
                    (param.name,
                     sum(('[%s]' % (soda_util.COORDS_IN_ORIG[d])
                          for d in range(len(param.size))), ''), '+'.join(
                              soda_util.COORDS_IN_ORIG[0:len(param.size)])))
    for d in param.size:
      printer.un_scope()
    printer.println()

  params = []
  for name in stencil.input_names + stencil.output_names + stencil.param_names:
    params.append(f'{data_fmt[name]}.data()')
    params.extend((extent_fmt[name], stride_fmt[name], min_fmt[name]))
  params.append('xclbin')
  printer.print_func(name=f'soda::app::{stencil.app_name}',
                     params=params,
                     align=0,
                     suffix=';')
  printer.println()

  printer.printlns('int error_count = 0;', '')

  for tensor in stencil.chronological_tensors:
    if tensor.is_input():
      continue
    logger.debug(f'emit code to produce {tensor.name}')
    printer.println('// produce %s, can be parallelized with -fopenmp' %
                    tensor.name)
    printer.println('#pragma omp parallel for', 0)
    stencil_window = core.get_overall_stencil_window(
        tuple(map(stencil.tensors.get, stencil.input_names))
        if tensor.is_output else tensor.parents.values(), tensor)
    stencil_dim = core.get_stencil_dim(stencil_window)
    output_idx = core.get_stencil_window_offset(stencil_window)
    for d in range(0, stencil.dim):
      dim = stencil.dim - d - 1
      printer.println(
          'for(int32_t {var} = {}; {var} < dims[{}] - {}; ++{var})'.format(
              max(0, output_idx[dim]),
              dim,
              max(0, stencil_dim[dim] - output_idx[dim] - 1),
              var=soda_util.COORDS_IN_ORIG[dim]))
      printer.do_scope()

    def mutate_load_for_host(obj, args):
      if isinstance(obj, ir.Ref):
        if obj.name in stencil.param_names:
          return ir.make_var(
              f'{data_fmt[obj.name]}%s' % ''.join('[%d]' % _ for _ in obj.idx),
              haoda_type=obj.haoda_type,
          )
        return ir.make_var(
            f'{data_fmt[obj.name]}[%s]' %
            (' + '.join(f'(%s%+d)*{stride_fmt[obj.name]}[%d]' %
                        (soda_util.COORDS_IN_ORIG[d],
                         obj.idx[d] - tensor.st_ref.idx[d], d)
                        for d in range(stencil.dim))),
            haoda_type=obj.haoda_type,
        )
      return obj

    def mutate_store_for_host(obj, args):
      if isinstance(obj, ir.Ref):
        if obj.name in stencil.output_names:
          return ir.make_var(
              '%s result_%s' % (obj.c_type, obj.name),
              haoda_type=obj.haoda_type,
          )
        return ir.make_var(
            f'{data_fmt[obj.name]}[%s]' %
            '+'.join(f'%s*{stride_fmt[obj.name]}[%d]' %
                     (soda_util.COORDS_IN_ORIG[d], d)
                     for d in range(stencil.dim)),
            haoda_type=obj.haoda_type,
        )
      return obj

    for let in tensor.lets:
      printer.printlns(
          '// let {} {} = {}'.format(let.haoda_type, let.name, let.expr),
          ('const {} {} = {};'.format(
              let.c_type, let.name,
              let.expr.visit(mutate_load_for_host).c_expr)),
      )
    printer.printlns(
        '// {} = {}'.format(tensor.st_ref, tensor.expr),
        ('{} = {};'.format(tensor.st_ref.visit(mutate_store_for_host),
                           tensor.expr.visit(mutate_load_for_host).c_expr)),
    )
    if tensor.is_output():
      run_result = f'{data_fmt[tensor.name]}[%s]' % (' + '.join(
          f'%s * {stride_fmt[tensor.name]}[%d]' %
          (soda_util.COORDS_IN_ORIG[d], d) for d in range(stencil.dim)))
      printer.printlns(
          '%s val_fpga = %s;' % (tensor.c_type, run_result),
          '%s val_cpu = result_%s;' % (tensor.c_type, tensor.name),
      )
      if tensor.haoda_type.is_float:
        printer.println('double threshold = 0.00001;')
        with printer.if_('nullptr != getenv("THRESHOLD")'):
          printer.println('threshold = atof(getenv("THRESHOLD"));')
        printer.println('threshold *= threshold;')
        with printer.if_(
            'double(val_fpga-val_cpu) * double(val_fpga-val_cpu) > '
            'threshold && '
            'double(val_fpga-val_cpu) * double(val_fpga-val_cpu) / '
            '(double(val_cpu) * double(val_cpu)) > threshold'):
          params = (', '.join(['%d'] * stencil.dim),
                    ', '.join(soda_util.COORDS_IN_ORIG[:stencil.dim]))
          printer.printlns(
              ('fprintf(stderr, "%%lf != %%lf @(%s)\\n", double'
               '(val_fpga), double(val_cpu), %s);' % params),
              '++error_count;',
          )
      else:
        with printer.if_('val_fpga != val_cpu'):
          params = (', '.join(['%d'] * stencil.dim),
                    ', '.join(soda_util.COORDS_IN_ORIG[:stencil.dim]))
          with printer.if_('error_count < 10'):
            printer.println('fprintf(stderr, "%%ld != %%ld @(%s)\\n", int64_t'
                            '(val_fpga), int64_t(val_cpu), %s);' % params)
          printer.println('++error_count;')
    for d in range(0, stencil.dim):
      printer.un_scope()
    printer.println()

  printer.printlns(
      r'clog << "INFO: " << (error_count ? "FAIL" : "PASS") << "!\n";',
      '',
  )

  printer.println('return error_count ? 1 : 0;')
  printer.un_scope()
  printer.println('#endif  // SODA_TEST_MAIN')


def print_code(stencil: core.Stencil, host_file: TextIO) -> None:
  logger.info('generate host source code as %s' % host_file.name)
  printer = util.CppPrinter(host_file)
  print_define = lambda key, value: util.print_define(printer, key, value)

  print_header(printer)

  if stencil.preserve_border:
    overall_stencil_window = core.get_overall_stencil_window(
        stencil.output.parent.preserve_border_from(), stencil.output)
  else:
    overall_stencil_window = core.get_overall_stencil_window(
        map(stencil.tensors.get, stencil.input_names),
        stencil.tensors[stencil.output_names[0]])

  overall_stencil_distance = core.get_stencil_distance(overall_stencil_window,
                                                       stencil.tile_size)

  stencil_offset = overall_stencil_distance - soda_util.serialize(
      core.get_stencil_window_offset(overall_stencil_window), stencil.tile_size)
  if stencil.preserve_border:
    stencil_offset *= stencil.iterate

  overall_stencil_distance = max(overall_stencil_distance, stencil_offset)

  all_stmts = stencil.input_stmts + stencil.output_stmts + stencil.param_stmts

  printer.printlns(
      'namespace soda {',
      'namespace app {',
      '// app-specific constants',
      *(f'constexpr int {STENCIL_DIM_FMT[i]} = {d};'
        for i, d in enumerate(core.get_stencil_dim(overall_stencil_window))),
      f'constexpr int kStencilDistance = {overall_stencil_distance};',
      *(f'constexpr int {WIDTH_FMT[x.name]} = {x.width_in_bits};'
        for x in all_stmts),
      '',
      '// type alias',
      *(f'using {TYPE_FMT[x.name]} = {x.c_type};' for x in all_stmts),
      '',
  )
  print_func(printer, stencil)
  printer.printlns(
      '} // namespace app',
      '} // namespace soda',
      '',
  )

  print_test(printer, stencil)



"""Stencil with Optimized Dataflow Architecture.

A stencil compiler that takes a high-level domain-specific language (DSL) as
input, applies parallelization, communication reuse, and computation reuse, and
generate optimized FPGA accelerators.

See:
https://github.com/Blaok/soda
"""

from setuptools import find_packages, setup

with open('../README.md', encoding='utf-8') as f:
  long_description = f.read()

setup(
    name='sodac',
    version='0.0.20200507.dev2',
    description='Stencil with optimized dataflow architecture',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Blaok/soda',
    author='Blaok Chi',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: System :: Hardware',
    ],
    packages=find_packages(exclude=('tests', 'tests.*')),
    python_requires='>=3.6',
    install_requires=[
        'cached_property',
        'haoda>=0.0.20200505.dev1',
        'pulp',
        'textx',
        'toposort',
    ],
    entry_points={
        'console_scripts': ['sodac=soda.sodac:main'],
    },
)



#!/usr/bin/python3
import argparse
import logging
import os
import sys

import textx

from haoda import util
from soda import core, grammar
from soda.codegen.frt import core as frt
from soda.codegen.intel import opencl as iocl
from soda.codegen.xilinx import opencl as xocl
from soda.model import xilinx as model
from soda.optimization import args as opt_args

logging.basicConfig(level=logging.WARNING,
                    format='%(levelname)s:%(name)s:%(lineno)d: %(message)s')
logger = logging.getLogger().getChild(os.path.basename(__file__))

def main():
  parser = argparse.ArgumentParser(
    prog='sodac',
    description='Stencil with Optimized Dataflow Architecture '
                '(SODA) compiler')
  parser.add_argument('--verbose', '-v',
                      action='count',
                      dest='verbose',
                      help='increase verbosity')
  parser.add_argument('--quiet', '-q',
                      action='count',
                      dest='quiet',
                      help='decrease verbosity')
  parser.add_argument('--recursion-limit',
                      type=int,
                      dest='recursion_limit',
                      help='override Python recursion limit')
  parser.add_argument('--burst-width',
                      type=int,
                      dest='burst_width',
                      help='override burst width')
  parser.add_argument('--unroll-factor',
                      type=int,
                      metavar='UNROLL_FACTOR',
                      dest='unroll_factor',
                      help='override unroll factor')
  parser.add_argument('--replication-factor',
                      type=int,
                      metavar='REPLICATION_FACTOR',
                      dest='replication_factor',
                      help='override replication factor')
  parser.add_argument('--tile-size',
                      type=int,
                      nargs='+',
                      metavar='TILE_SIZE',
                      dest='tile_size',
                      help='override tile size; '
                           '0 means no overriding on that dimension')
  parser.add_argument('--dram-in',
                      type=str,
                      dest='dram_in',
                      help='override DRAM configuration for input')
  parser.add_argument('--dram-out',
                      type=str,
                      dest='dram_out',
                      help='override DRAM configuration for output')
  parser.add_argument('--iterate',
            type=int,
            metavar='#ITERATION',
            dest='iterate',
            help='override iterate directive; '
            'repeat execution multiple times iteratively')
  parser.add_argument('--border',
            type=str,
            metavar='(ignore|preserve)',
            dest='border',
            help='override border handling strategy')
  parser.add_argument('--cluster',
            type=str,
            metavar='(none|fine|coarse|full)',
            dest='cluster',
            help='module clustering level, `none` generates '
               'standalone compute / forward modules, `fine` '
               'fuses forwarders into compute modules, `coarse` '
               'fuses each stage together, `full` fuses '
               'everything together')
  parser.add_argument(type=str,
            dest='soda_src',
            metavar='file',
            help='soda source code')

  xocl.add_arguments(parser.add_argument_group('Xilinx OpenCL backend'))
  iocl.add_arguments(parser.add_argument_group('Intel OpenCL backend'))
  frt.add_arguments(parser.add_argument_group('FPGA runtime backend'))
  opt_args.add_arguments(parser.add_argument_group('SODA optimizations'))

  parser.add_argument('--model-file',
            type=str,
            dest='model_file',
            metavar='file',
            help='resource model specified as json file')
  parser.add_argument('--estimation-file',
            type=str,
            dest='estimation_file',
            metavar='file',
            help='report resource and performance estimation as '
               'json file')

  args = parser.parse_args()
  verbose = 0 if args.verbose is None else args.verbose
  quiet = 0 if args.quiet is None else args.quiet
  logging_level = (quiet-verbose)*10+logging.getLogger().getEffectiveLevel()
  if logging_level > logging.CRITICAL:
    logging_level = logging.CRITICAL
  if logging_level < logging.DEBUG:
    logging_level = logging.DEBUG
  logging.getLogger().setLevel(logging_level)
  logger.info('set log level to %s', logging.getLevelName(logging_level))
  # TODO: check tile size

  if args.recursion_limit is not None:
    sys_recursion_limit = sys.getrecursionlimit()
    if sys_recursion_limit > args.recursion_limit:
      logger.warning(
        'Python system recursion limit (%d) > specified value (%d); '
        'the latter will be ignored', sys_recursion_limit, args.recursion_limit)
    else:
      sys.setrecursionlimit(args.recursion_limit)
      logger.warning('Python recursion limit is set to %d',
                     sys.getrecursionlimit())

  soda_mm = textx.metamodel_from_str(grammar.GRAMMAR, classes=grammar.CLASSES)
  logger.info('build metamodel')
  try:
    if args.soda_src == '-':
      soda_file_name = sys.stdin.name
      soda_model = soda_mm.model_from_str(sys.stdin.read())
    else:
      with open(args.soda_src, 'r') as soda_file:
        soda_model = soda_mm.model_from_str(soda_file.read())
        soda_file_name = soda_file.name
    logger.info('%s parsed as soda file', soda_file_name)
    logger.debug('soda program parsed:\n  %s',
                 str(soda_model).replace('\n', '\n  '))

    tile_size = []
    for dim in range(soda_model.dim-1):
      if (args.tile_size is not None and
          dim < len(args.tile_size) and
          args.tile_size[dim] > 0):
        tile_size.append(args.tile_size[dim])
      else:
        tile_size.append(soda_model.tile_size[dim])
    tile_size.append(0)

    if args.replication_factor is None:
      if args.unroll_factor is not None:
        unroll_factor = args.unroll_factor
      else:
        unroll_factor = soda_model.unroll_factor
      replication_factor = 1
    else:
      unroll_factor = args.replication_factor
      replication_factor = args.replication_factor

    stencil = core.Stencil(
      burst_width=args.burst_width if args.burst_width is not None
                     else soda_model.burst_width,
      border=args.border if args.border is not None
                 else soda_model.border,
      iterate=args.iterate if args.iterate is not None
                 else soda_model.iterate,
      cluster=args.cluster if args.cluster is not None
                 else soda_model.cluster,
      dram_in=args.dram_in,
      dram_out=args.dram_out,
      app_name=soda_model.app_name,
      input_stmts=soda_model.input_stmts,
      param_stmts=soda_model.param_stmts,
      local_stmts=soda_model.local_stmts,
      output_stmts=soda_model.output_stmts,
      dim=soda_model.dim,
      tile_size=tile_size,
      unroll_factor=unroll_factor,
      replication_factor=replication_factor,
      optimizations=opt_args.get_kwargs(args),
    )

    logger.debug('stencil obtained: %s', stencil)

    xocl.print_code(stencil, args, parser)
    iocl.print_code(stencil, args)
    frt.print_code(stencil, args)

    if args.estimation_file is not None:
      if args.model_file is None:
        if args.soda_src.endswith('.soda'):
          model_file = args.soda_src[:-len('.soda')]+'_model.json'
        else:
          logger.fatal('cannot find resource model file')
          sys.exit(1)
      else:
        model_file = args.model_file

      def print_estimation():
        def print_estimation():
          model.print_estimation(stencil,
                       model_file,
                       estimation_file)
        if args.estimation_file == '-':
          estimation_file = sys.stdout
          print_estimation()
        else:
          with open(args.estimation_file, 'w') as estimation_file:
            print_estimation()
      if model_file == '-':
        model_file = sys.stdin
        print_estimation()
      else:
        with open(model_file) as model_file:
          print_estimation()


  except textx.exceptions.TextXSyntaxError as e:
    logger.error(e)
    sys.exit(1)
  except util.SemanticError as e:
    logger.error(e)
    sys.exit(1)
  except util.SemanticWarn as w:
    logger.warning(w)

if __name__ == '__main__':
  main()



import argparse
from typing import Dict


def add_arguments(parser: argparse.ArgumentParser) -> None:
  parser.add_argument('--inline',
                      type=str,
                      metavar='(yes|no)',
                      dest='inline',
                      nargs='?',
                      const='yes',
                      default='no')
  parser.add_argument(
      '--computation-reuse',
      type=str,
      metavar=
      '(yes|no|greedy|optimal|glore|built-in|built-in:greedy|built-in:optimal)',
      dest='computation_reuse',
      nargs='?',
      const='yes',
      default='no',
      help='enable computation reuse or not')


def get_kwargs(args: argparse.Namespace) -> Dict[str, str]:
  optimizations = {}
  if args.computation_reuse != 'no':
    optimizations['computation-reuse'] = args.computation_reuse
  if args.inline != 'no':
    optimizations['inline'] = args.inline
  return optimizations
